#
# Generated by TigerShark.tools.convertPyX12 on 2012-10-03 14:38:27.454816
#
from tigershark.X12.parse import Message, Loop, Segment, Composite, Element, Properties
parsed_270_HEADER = Loop( u'HEADER', Properties(looptype=u'wrapper',repeat=u'1',pos=u'015',req_sit=u'R',desc=u'Table 1 - Header'), 
Segment( u'BHT', Properties(syntax='',req_sit=u'R',repeat=u'1',pos=u'020',desc=u'Beginning of Hierarchical Transaction'),
  Element( u'BHT01', Properties(desc=u'Hierarchical Structure Code', req_sit=u'R', data_type=(u'ID',u'4',u'4'), position=1,
    codes=[u'0022'] ) ),
  Element( u'BHT02', Properties(desc=u'Transaction Set Purpose Code', req_sit=u'R', data_type=(u'ID',u'2',u'2'), position=2,
    codes=[u'01', u'13', u'36'] ) ),
  Element( u'BHT03', Properties(desc=u'Reference Identification', req_sit=u'S', data_type=(u'AN',u'1',u'50'), position=3,
    codes=[] ) ),
  Element( u'BHT04', Properties(desc=u'Date', req_sit=u'R', data_type=(u'DT',u'8',u'8'), position=4,
    codes=[] ) ),
  Element( u'BHT05', Properties(desc=u'Time', req_sit=u'R', data_type=(u'TM',u'4',u'8'), position=5,
    codes=[] ) ),
  Element( u'BHT06', Properties(desc=u'Transaction Type Code', req_sit=u'S', data_type=(u'ID',u'2',u'2'), position=6,
    codes=[u'RT', u'RU'] ) ),
),
)
parsed_270_2100A = Loop( u'2100A', Properties(looptype='',repeat=u'1',pos=u'030',req_sit=u'R',desc=u'Information Source Name'), 
Segment( u'NM1', Properties(syntax='',req_sit=u'R',repeat=u'1',pos=u'030',desc=u'Information Source Name'),
  Element( u'NM101', Properties(desc=u'Entity Identifier Code', req_sit=u'R', data_type=(u'ID',u'2',u'3'), position=1,
    codes=[u'2B', u'36', u'GP', u'P5', u'PR'] ) ),
  Element( u'NM102', Properties(desc=u'Entity Type Qualifier', req_sit=u'R', data_type=(u'ID',u'1',u'1'), position=2,
    codes=[u'1', u'2'] ) ),
  Element( u'NM103', Properties(desc=u'Name Last or Organization Name', req_sit=u'S', data_type=(u'AN',u'1',u'60'), position=3,
    codes=[] ) ),
  Element( u'NM104', Properties(desc=u'Name First', req_sit=u'S', data_type=(u'AN',u'1',u'35'), position=4,
    codes=[] ) ),
  Element( u'NM105', Properties(desc=u'Name Middle', req_sit=u'S', data_type=(u'AN',u'1',u'25'), position=5,
    codes=[] ) ),
  Element( u'NM106', Properties(desc=u'Name Prefix', req_sit=u'N', data_type=(u'AN',u'1',u'10'), position=6,
    codes=[] ) ),
  Element( u'NM107', Properties(desc=u'Name Suffix', req_sit=u'S', data_type=(u'AN',u'1',u'10'), position=7,
    codes=[] ) ),
  Element( u'NM108', Properties(desc=u'Identification Code Qualifier', req_sit=u'R', data_type=(u'ID',u'1',u'2'), position=8,
    codes=[u'24', u'46', u'FI', u'NI', u'PI', u'XV', u'XX'] ) ),
  Element( u'NM109', Properties(desc=u'Identification Code', req_sit=u'R', data_type=(u'AN',u'2',u'80'), position=9,
    codes=[] ) ),
  Element( u'NM110', Properties(desc=u'Entity Relationship Code', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=10,
    codes=[] ) ),
  Element( u'NM111', Properties(desc=u'Entity Identifier Code', req_sit=u'N', data_type=(u'ID',u'2',u'3'), position=11,
    codes=[] ) ),
),
)
parsed_270_2100B = Loop( u'2100B', Properties(looptype='',repeat=u'1',pos=u'030',req_sit=u'R',desc=u'Information Receiver Name'), 
Segment( u'NM1', Properties(syntax='',req_sit=u'R',repeat=u'1',pos=u'030',desc=u'Information Receiver Name'),
  Element( u'NM101', Properties(desc=u'Entity Identifier Code', req_sit=u'R', data_type=(u'ID',u'2',u'3'), position=1,
    codes=[u'1P', u'2B', u'36', u'80', u'FA', u'GP', u'P5', u'PR'] ) ),
  Element( u'NM102', Properties(desc=u'Entity Type Qualifier', req_sit=u'R', data_type=(u'ID',u'1',u'1'), position=2,
    codes=[u'1', u'2'] ) ),
  Element( u'NM103', Properties(desc=u'Name Last or Organization Name', req_sit=u'S', data_type=(u'AN',u'1',u'60'), position=3,
    codes=[] ) ),
  Element( u'NM104', Properties(desc=u'Name First', req_sit=u'S', data_type=(u'AN',u'1',u'35'), position=4,
    codes=[] ) ),
  Element( u'NM105', Properties(desc=u'Name Middle', req_sit=u'S', data_type=(u'AN',u'1',u'25'), position=5,
    codes=[] ) ),
  Element( u'NM106', Properties(desc=u'Name Prefix', req_sit=u'N', data_type=(u'AN',u'1',u'10'), position=6,
    codes=[] ) ),
  Element( u'NM107', Properties(desc=u'Name Suffix', req_sit=u'S', data_type=(u'AN',u'1',u'10'), position=7,
    codes=[] ) ),
  Element( u'NM108', Properties(desc=u'Identification Code Qualifier', req_sit=u'R', data_type=(u'ID',u'1',u'2'), position=8,
    codes=[u'24', u'34', u'FI', u'PI', u'PP', u'SV', u'XV', u'XX'] ) ),
  Element( u'NM109', Properties(desc=u'Identification Code', req_sit=u'R', data_type=(u'AN',u'2',u'80'), position=9,
    codes=[] ) ),
  Element( u'NM110', Properties(desc=u'Entity Relationship Code', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=10,
    codes=[] ) ),
  Element( u'NM111', Properties(desc=u'Entity Identifier Code', req_sit=u'N', data_type=(u'ID',u'2',u'3'), position=11,
    codes=[] ) ),
),
Segment( u'REF', Properties(syntax='',req_sit=u'S',repeat=u'9',pos=u'040',desc=u'Information Receiver Additional Identification'),
  Element( u'REF01', Properties(desc=u'Reference Identification Qualifier', req_sit=u'R', data_type=(u'ID',u'2',u'3'), position=1,
    codes=[u'0B', u'1C', u'1D', u'1J', u'4A', u'CT', u'EL', u'EO', u'JD', u'N5', u'N7', u'Q4', u'SY', u'TJ', u'HPI'] ) ),
  Element( u'REF02', Properties(desc=u'Reference Identification', req_sit=u'R', data_type=(u'AN',u'1',u'50'), position=2,
    codes=[] ) ),
  Element( u'REF03', Properties(desc=u'Description', req_sit=u'S', data_type=(u'AN',u'1',u'80'), position=3,
    codes=[] ) ),
  Composite( u'C040', Properties(req_sit=u'N',refdes='',seq=u'04',desc=u'Reference Identifier'),
  ),
),
Segment( u'N3', Properties(syntax='',req_sit=u'S',repeat=u'1',pos=u'060',desc=u'Information Receiver Address'),
  Element( u'N301', Properties(desc=u'Address Information', req_sit=u'R', data_type=(u'AN',u'1',u'55'), position=1,
    codes=[] ) ),
  Element( u'N302', Properties(desc=u'Address Information', req_sit=u'S', data_type=(u'AN',u'1',u'55'), position=2,
    codes=[] ) ),
),
Segment( u'N4', Properties(syntax='',req_sit=u'S',repeat=u'1',pos=u'070',desc=u'Information Receiver City/State/Zip Code'),
  Element( u'N401', Properties(desc=u'City Name', req_sit=u'R', data_type=(u'AN',u'2',u'30'), position=1,
    codes=[] ) ),
  Element( u'N402', Properties(desc=u'State or Province Code', req_sit=u'R', data_type=(u'ID',u'2',u'2'), position=2,
    codes=[] ) ),
  Element( u'N403', Properties(desc=u'Postal Code', req_sit=u'R', data_type=(u'ID',u'3',u'15'), position=3,
    codes=[] ) ),
  Element( u'N404', Properties(desc=u'Country Code', req_sit=u'S', data_type=(u'ID',u'2',u'3'), position=4,
    codes=[] ) ),
  Element( u'N405', Properties(desc=u'Location Qualifier', req_sit=u'N', data_type=(u'ID',u'1',u'2'), position=5,
    codes=[] ) ),
  Element( u'N406', Properties(desc=u'Location Identifier', req_sit=u'N', data_type=(u'AN',u'1',u'30'), position=6,
    codes=[] ) ),
),
Segment( u'PER', Properties(syntax='',req_sit=u'S',repeat=u'3',pos=u'080',desc=u'Information Receiver Contact Information'),
  Element( u'PER01', Properties(desc=u'Contact Function Code', req_sit=u'R', data_type=(u'ID',u'2',u'2'), position=1,
    codes=[u'IC'] ) ),
  Element( u'PER02', Properties(desc=u'Name', req_sit=u'S', data_type=(u'AN',u'1',u'60'), position=2,
    codes=[] ) ),
  Element( u'PER03', Properties(desc=u'Communication Number Qualifier', req_sit=u'S', data_type=(u'ID',u'2',u'2'), position=3,
    codes=[u'ED', u'EM', u'FX', u'TE'] ) ),
  Element( u'PER04', Properties(desc=u'Communication Number', req_sit=u'S', data_type=(u'AN',u'1',u'256'), position=4,
    codes=[] ) ),
  Element( u'PER05', Properties(desc=u'Communication Number Qualifier', req_sit=u'S', data_type=(u'ID',u'2',u'2'), position=5,
    codes=[u'ED', u'EM', u'EX', u'FX', u'TE'] ) ),
  Element( u'PER06', Properties(desc=u'Communication Number', req_sit=u'S', data_type=(u'AN',u'1',u'256'), position=6,
    codes=[] ) ),
  Element( u'PER07', Properties(desc=u'Communication Number Qualifier', req_sit=u'S', data_type=(u'ID',u'2',u'2'), position=7,
    codes=[u'ED', u'EM', u'EX', u'FX', u'TE'] ) ),
  Element( u'PER08', Properties(desc=u'Communication Number', req_sit=u'S', data_type=(u'AN',u'1',u'256'), position=8,
    codes=[] ) ),
  Element( u'PER09', Properties(desc=u'Contact Inquiry Reference', req_sit=u'N', data_type=(u'AN',u'1',u'20'), position=9,
    codes=[] ) ),
),
Segment( u'PRV', Properties(syntax='',req_sit=u'S',repeat=u'1',pos=u'090',desc=u'Information Receiver Provider Information'),
  Element( u'PRV01', Properties(desc=u'Provider Code', req_sit=u'R', data_type=(u'ID',u'1',u'3'), position=1,
    codes=[u'AD', u'AT', u'BI', u'CO', u'CV', u'H', u'HH', u'LA', u'OT', u'P1', u'P2', u'PC', u'PE', u'R', u'RF', u'SB', u'SK', u'SU'] ) ),
  Element( u'PRV02', Properties(desc=u'Reference Identification Qualifier', req_sit=u'R', data_type=(u'ID',u'2',u'3'), position=2,
    codes=[u'ZZ'] ) ),
  Element( u'PRV03', Properties(desc=u'Reference Identification', req_sit=u'R', data_type=(u'AN',u'1',u'50'), position=3,
    codes=[] ) ),
  Element( u'PRV04', Properties(desc=u'State or Province Code', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=4,
    codes=[] ) ),
  Composite( u'C035', Properties(req_sit=u'N',refdes='',seq=u'05',desc=u'Provider Specialty Information'),
  ),
  Element( u'PRV06', Properties(desc=u'Provider Organization Code', req_sit=u'N', data_type=(u'ID',u'3',u'3'), position=6,
    codes=[] ) ),
),
)
parsed_270_2110C = Loop( u'2110C', Properties(looptype='',repeat=u'99',pos=u'130',req_sit=u'S',desc=u'Subscriber Eligibility or Benefit Inquiry Information'), 
Segment( u'EQ', Properties(syntax='',req_sit=u'R',repeat=u'1',pos=u'130',desc=u'Subscriber Eligibility or Benefit Inquiry Information'),
  Element( u'EQ01', Properties(desc=u'Service Type Code', req_sit=u'S', data_type=(u'ID',u'1',u'2'), position=1,
    codes=[u'1', u'2', u'3', u'4', u'5', u'6', u'7', u'8', u'9', u'10', u'11', u'12', u'13', u'14', u'15', u'16', u'17', u'18', u'19', u'20', u'21', u'22', u'23', u'24', u'25', u'26', u'27', u'28', u'30', u'32', u'33', u'34', u'35', u'36', u'37', u'38', u'39', u'40', u'41', u'42', u'43', u'44', u'45', u'46', u'47', u'48', u'49', u'50', u'51', u'52', u'53', u'54', u'55', u'56', u'57', u'58', u'59', u'60', u'61', u'62', u'63', u'64', u'65', u'66', u'67', u'68', u'69', u'70', u'71', u'72', u'73', u'74', u'75', u'76', u'77', u'78', u'79', u'80', u'81', u'82', u'83', u'84', u'85', u'86', u'87', u'88', u'89', u'90', u'91', u'92', u'93', u'94', u'95', u'96', u'97', u'98', u'99', u'A0', u'A1', u'A2', u'A3', u'A4', u'A5', u'A6', u'A7', u'A8', u'A9', u'AA', u'AB', u'AC', u'AD', u'AE', u'AF', u'AG', u'AH', u'AI', u'AJ', u'AK', u'AL', u'AM', u'AN', u'AO', u'AQ', u'AR', u'BA', u'BB', u'BC', u'BD', u'BE', u'BF', u'BG', u'BH', u'BI', u'BJ', u'BK', u'BL', u'BM', u'BN', u'BP', u'BQ', u'BR', u'BS'] ) ),
  Composite( u'C003', Properties(req_sit=u'S',refdes='',seq=u'02',desc=u'Composite Medical Procedure Identifier'),
    Element( u'EQ02-01', Properties(desc=u'Product/Service ID Qualifier', req_sit=u'R', data_type=(u'ID',u'2',u'2'), position=0,
      codes=[u'AD', u'CJ', u'HC', u'ID', u'IV', u'N4', u'ZZ'] ) ),
    Element( u'EQ02-02', Properties(desc=u'Product/Service ID', req_sit=u'R', data_type=(u'AN',u'1',u'48'), position=1,
      codes=[] ) ),
    Element( u'EQ02-03', Properties(desc=u'Procedure Modifier', req_sit=u'S', data_type=(u'AN',u'2',u'2'), position=2,
      codes=[] ) ),
    Element( u'EQ02-04', Properties(desc=u'Procedure Modifier', req_sit=u'S', data_type=(u'AN',u'2',u'2'), position=3,
      codes=[] ) ),
    Element( u'EQ02-05', Properties(desc=u'Procedure Modifier', req_sit=u'S', data_type=(u'AN',u'2',u'2'), position=4,
      codes=[] ) ),
    Element( u'EQ02-06', Properties(desc=u'Procedure Modifier', req_sit=u'S', data_type=(u'AN',u'2',u'2'), position=5,
      codes=[] ) ),
    Element( u'EQ02-07', Properties(desc=u'Description', req_sit=u'N', data_type=(u'AN',u'1',u'80'), position=6,
      codes=[] ) ),
  ),
  Element( u'EQ03', Properties(desc=u'Coverage Level Code', req_sit=u'S', data_type=(u'ID',u'3',u'3'), position=3,
    codes=[u'CHD', u'DEP', u'ECH', u'EMP', u'ESP', u'FAM', u'IND', u'SPC', u'SPO'] ) ),
  Element( u'EQ04', Properties(desc=u'Insurance Type Code', req_sit=u'S', data_type=(u'ID',u'1',u'3'), position=4,
    codes=[u'AP', u'C1', u'CO', u'GP', u'HM', u'HN', u'IP', u'MA', u'MB', u'MC', u'PR', u'PS', u'SP', u'WC'] ) ),
),
Segment( u'AMT', Properties(syntax='',req_sit=u'S',repeat=u'1',pos=u'135',desc=u'Subscriber Spend Down Amount'),
  Element( u'AMT01', Properties(desc=u'Amount Qualifier Code', req_sit=u'R', data_type=(u'ID',u'1',u'3'), position=1,
    codes=[u'R'] ) ),
  Element( u'AMT02', Properties(desc=u'Monetary Amount', req_sit=u'R', data_type=(u'R',u'1',u'18'), position=2,
    codes=[] ) ),
  Element( u'AMT03', Properties(desc=u'Credit/Debit Flag Code', req_sit=u'N', data_type=(u'ID',u'1',u'1'), position=3,
    codes=[] ) ),
),
Segment( u'III', Properties(syntax='',req_sit=u'S',repeat=u'10',pos=u'170',desc=u'Subscriber Eligibility or Benefit Additional Inquiry Information'),
  Element( u'III01', Properties(desc=u'Code List Qualifier Code', req_sit=u'R', data_type=(u'ID',u'1',u'3'), position=1,
    codes=[u'BF', u'BK', u'ZZ'] ) ),
  Element( u'III02', Properties(desc=u'Industry Code', req_sit=u'R', data_type=(u'AN',u'1',u'30'), position=2,
    codes=[] ) ),
  Element( u'III03', Properties(desc=u'Code Category', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=3,
    codes=[] ) ),
  Element( u'III04', Properties(desc=u'Free-form Message Text', req_sit=u'N', data_type=(u'AN',u'1',u'264'), position=4,
    codes=[] ) ),
  Element( u'III05', Properties(desc=u'Quantity', req_sit=u'N', data_type=(u'R',u'1',u'15'), position=5,
    codes=[] ) ),
  Composite( u'C001', Properties(req_sit=u'N',refdes='',seq=u'06',desc=u'Composite Unit of Measure'),
  ),
  Element( u'III07', Properties(desc=u'Surface/Layer/Position Code', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=7,
    codes=[] ) ),
  Element( u'III08', Properties(desc=u'Surface/Layer/Position Code', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=8,
    codes=[] ) ),
  Element( u'III09', Properties(desc=u'Surface/Layer/Position Code', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=9,
    codes=[] ) ),
),
Segment( u'REF', Properties(syntax='',req_sit=u'S',repeat=u'1',pos=u'190',desc=u'Subscriber Additional Information'),
  Element( u'REF01', Properties(desc=u'Reference Identification Qualifier', req_sit=u'R', data_type=(u'ID',u'2',u'3'), position=1,
    codes=[u'9F', u'G1'] ) ),
  Element( u'REF02', Properties(desc=u'Reference Identification', req_sit=u'R', data_type=(u'AN',u'1',u'50'), position=2,
    codes=[] ) ),
  Element( u'REF03', Properties(desc=u'Description', req_sit=u'N', data_type=(u'AN',u'1',u'80'), position=3,
    codes=[] ) ),
  Composite( u'C040', Properties(req_sit=u'N',refdes='',seq=u'04',desc=u'Reference Identifier'),
  ),
),
Segment( u'DTP', Properties(syntax='',req_sit=u'S',repeat=u'1',pos=u'200',desc=u'Subscriber Eligibility/Benefit Date'),
  Element( u'DTP01', Properties(desc=u'Date/Time Qualifier', req_sit=u'R', data_type=(u'ID',u'3',u'3'), position=1,
    codes=[u'307', u'435', u'472'] ) ),
  Element( u'DTP02', Properties(desc=u'Date Time Period Format Qualifier', req_sit=u'R', data_type=(u'ID',u'2',u'3'), position=2,
    codes=[u'D8', u'RD8'] ) ),
  Element( u'DTP03', Properties(desc=u'Date Time Period', req_sit=u'R', data_type=(u'AN',u'1',u'35'), position=3,
    codes=[] ) ),
),
)
parsed_270_2100C = Loop( u'2100C', Properties(looptype='',repeat=u'1',pos=u'030',req_sit=u'R',desc=u'Subscriber Name'), 
Segment( u'NM1', Properties(syntax='',req_sit=u'R',repeat=u'1',pos=u'030',desc=u'Subscriber Name'),
  Element( u'NM101', Properties(desc=u'Entity Identifier Code', req_sit=u'R', data_type=(u'ID',u'2',u'3'), position=1,
    codes=[u'IL'] ) ),
  Element( u'NM102', Properties(desc=u'Entity Type Qualifier', req_sit=u'R', data_type=(u'ID',u'1',u'1'), position=2,
    codes=[u'1'] ) ),
  Element( u'NM103', Properties(desc=u'Name Last or Organization Name', req_sit=u'S', data_type=(u'AN',u'1',u'60'), position=3,
    codes=[] ) ),
  Element( u'NM104', Properties(desc=u'Name First', req_sit=u'S', data_type=(u'AN',u'1',u'35'), position=4,
    codes=[] ) ),
  Element( u'NM105', Properties(desc=u'Name Middle', req_sit=u'S', data_type=(u'AN',u'1',u'25'), position=5,
    codes=[] ) ),
  Element( u'NM106', Properties(desc=u'Name Prefix', req_sit=u'N', data_type=(u'AN',u'1',u'10'), position=6,
    codes=[] ) ),
  Element( u'NM107', Properties(desc=u'Name Suffix', req_sit=u'S', data_type=(u'AN',u'1',u'10'), position=7,
    codes=[] ) ),
  Element( u'NM108', Properties(desc=u'Identification Code Qualifier', req_sit=u'S', data_type=(u'ID',u'1',u'2'), position=8,
    codes=[u'MI', u'ZZ'] ) ),
  Element( u'NM109', Properties(desc=u'Identification Code', req_sit=u'S', data_type=(u'AN',u'2',u'80'), position=9,
    codes=[] ) ),
  Element( u'NM110', Properties(desc=u'Entity Relationship Code', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=10,
    codes=[] ) ),
  Element( u'NM111', Properties(desc=u'Entity Identifier Code', req_sit=u'N', data_type=(u'ID',u'2',u'3'), position=11,
    codes=[] ) ),
),
Segment( u'REF', Properties(syntax='',req_sit=u'S',repeat=u'9',pos=u'040',desc=u'Subscriber Additional Identification'),
  Element( u'REF01', Properties(desc=u'Reference Identification Qualifier', req_sit=u'R', data_type=(u'ID',u'2',u'3'), position=1,
    codes=[u'18', u'1L', u'1W', u'49', u'6P', u'A6', u'CT', u'EA', u'EJ', u'F6', u'GH', u'HJ', u'IG', u'N6', u'NQ', u'SY'] ) ),
  Element( u'REF02', Properties(desc=u'Reference Identification', req_sit=u'R', data_type=(u'AN',u'1',u'50'), position=2,
    codes=[] ) ),
  Element( u'REF03', Properties(desc=u'Description', req_sit=u'N', data_type=(u'AN',u'1',u'80'), position=3,
    codes=[] ) ),
  Composite( u'C040', Properties(req_sit=u'N',refdes='',seq=u'04',desc=u'Reference Identifier'),
  ),
),
Segment( u'N3', Properties(syntax='',req_sit=u'S',repeat=u'1',pos=u'060',desc=u'Subscriber Address'),
  Element( u'N301', Properties(desc=u'Address Information', req_sit=u'R', data_type=(u'AN',u'1',u'55'), position=1,
    codes=[] ) ),
  Element( u'N302', Properties(desc=u'Address Information', req_sit=u'S', data_type=(u'AN',u'1',u'55'), position=2,
    codes=[] ) ),
),
Segment( u'N4', Properties(syntax='',req_sit=u'S',repeat=u'1',pos=u'070',desc=u'Subscriber City/State/ZIP Code'),
  Element( u'N401', Properties(desc=u'City Name', req_sit=u'S', data_type=(u'AN',u'2',u'30'), position=1,
    codes=[] ) ),
  Element( u'N402', Properties(desc=u'State or Province Code', req_sit=u'S', data_type=(u'ID',u'2',u'2'), position=2,
    codes=[] ) ),
  Element( u'N403', Properties(desc=u'Postal Code', req_sit=u'S', data_type=(u'ID',u'3',u'15'), position=3,
    codes=[] ) ),
  Element( u'N404', Properties(desc=u'Country Code', req_sit=u'S', data_type=(u'ID',u'2',u'3'), position=4,
    codes=[] ) ),
  Element( u'N405', Properties(desc=u'Location Qualifier', req_sit=u'N', data_type=(u'ID',u'1',u'2'), position=5,
    codes=[] ) ),
  Element( u'N406', Properties(desc=u'Location Identifier', req_sit=u'N', data_type=(u'AN',u'1',u'30'), position=6,
    codes=[] ) ),
),
Segment( u'PRV', Properties(syntax='',req_sit=u'S',repeat=u'1',pos=u'090',desc=u'Provider Information'),
  Element( u'PRV01', Properties(desc=u'Provider Code', req_sit=u'R', data_type=(u'ID',u'1',u'3'), position=1,
    codes=[u'AD', u'AT', u'BI', u'CO', u'CV', u'H', u'HH', u'LA', u'OT', u'P1', u'P2', u'PC', u'PE', u'R', u'RF', u'SB', u'SK', u'SU'] ) ),
  Element( u'PRV02', Properties(desc=u'Reference Identification Qualifier', req_sit=u'R', data_type=(u'ID',u'2',u'3'), position=2,
    codes=[u'9K', u'D3', u'EI', u'HPI', u'SY', u'TJ', u'ZZ'] ) ),
  Element( u'PRV03', Properties(desc=u'Reference Identification', req_sit=u'R', data_type=(u'AN',u'1',u'50'), position=3,
    codes=[] ) ),
  Element( u'PRV04', Properties(desc=u'State or Province Code', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=4,
    codes=[] ) ),
  Composite( u'C035', Properties(req_sit=u'N',refdes='',seq=u'05',desc=u'Provider Specialty Information'),
  ),
  Element( u'PRV06', Properties(desc=u'Provider Organization Code', req_sit=u'N', data_type=(u'ID',u'3',u'3'), position=6,
    codes=[] ) ),
),
Segment( u'DMG', Properties(syntax='',req_sit=u'S',repeat=u'1',pos=u'100',desc=u'Subscriber Demographic Information'),
  Element( u'DMG01', Properties(desc=u'Date Time Period Format Qualifier', req_sit=u'S', data_type=(u'ID',u'2',u'3'), position=1,
    codes=[u'D8'] ) ),
  Element( u'DMG02', Properties(desc=u'Date Time Period', req_sit=u'S', data_type=(u'AN',u'1',u'35'), position=2,
    codes=[] ) ),
  Element( u'DMG03', Properties(desc=u'Gender Code', req_sit=u'S', data_type=(u'ID',u'1',u'1'), position=3,
    codes=[u'F', u'M'] ) ),
  Element( u'DMG04', Properties(desc=u'Marital Status Code', req_sit=u'N', data_type=(u'ID',u'1',u'1'), position=4,
    codes=[] ) ),
  Element( u'DMG05', Properties(desc=u'Race or Ethnicity Code', req_sit=u'N', data_type=(u'ID',u'1',u'1'), position=5,
    codes=[] ) ),
  Element( u'DMG06', Properties(desc=u'Citizenship Status Code', req_sit=u'N', data_type=(u'ID',u'1',u'2'), position=6,
    codes=[] ) ),
  Element( u'DMG07', Properties(desc=u'Country Code', req_sit=u'N', data_type=(u'ID',u'2',u'3'), position=7,
    codes=[] ) ),
  Element( u'DMG08', Properties(desc=u'Basis of Verification Code', req_sit=u'N', data_type=(u'ID',u'1',u'2'), position=8,
    codes=[] ) ),
  Element( u'DMG09', Properties(desc=u'Quantity', req_sit=u'N', data_type=(u'R',u'1',u'15'), position=9,
    codes=[] ) ),
),
Segment( u'INS', Properties(syntax='',req_sit=u'S',repeat=u'1',pos=u'110',desc=u'Subscriber Relationship'),
  Element( u'INS01', Properties(desc=u'Yes/No Condition or Response Code', req_sit=u'R', data_type=(u'ID',u'1',u'1'), position=1,
    codes=[u'Y'] ) ),
  Element( u'INS02', Properties(desc=u'Individual Relationship Code', req_sit=u'R', data_type=(u'ID',u'2',u'2'), position=2,
    codes=[u'18'] ) ),
  Element( u'INS03', Properties(desc=u'Maintenance Type Code', req_sit=u'N', data_type=(u'ID',u'3',u'3'), position=3,
    codes=[] ) ),
  Element( u'INS04', Properties(desc=u'Maintenance Reason Code', req_sit=u'N', data_type=(u'ID',u'2',u'3'), position=4,
    codes=[] ) ),
  Element( u'INS05', Properties(desc=u'Benefit Status Code', req_sit=u'N', data_type=(u'ID',u'1',u'1'), position=5,
    codes=[] ) ),
  Element( u'INS06', Properties(desc=u'Medicare Plan Code', req_sit=u'N', data_type=(u'ID',u'1',u'1'), position=6,
    codes=[] ) ),
  Element( u'INS07', Properties(desc=u'Consolidated Omnibus Budget Reconciliation Act (COBRA) Qualifying', req_sit=u'N', data_type=(u'ID',u'1',u'2'), position=7,
    codes=[] ) ),
  Element( u'INS08', Properties(desc=u'Employment Status Code', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=8,
    codes=[] ) ),
  Element( u'INS09', Properties(desc=u'Student Status Code', req_sit=u'N', data_type=(u'ID',u'1',u'1'), position=9,
    codes=[] ) ),
  Element( u'INS10', Properties(desc=u'Yes/No Condition or Response Code', req_sit=u'N', data_type=(u'ID',u'1',u'1'), position=10,
    codes=[] ) ),
  Element( u'INS11', Properties(desc=u'Date Time Period Format Qualifier', req_sit=u'N', data_type=(u'ID',u'2',u'3'), position=11,
    codes=[] ) ),
  Element( u'INS12', Properties(desc=u'Date Time Period', req_sit=u'N', data_type=(u'AN',u'1',u'35'), position=12,
    codes=[] ) ),
  Element( u'INS13', Properties(desc=u'Confidentiality Code', req_sit=u'N', data_type=(u'ID',u'1',u'1'), position=13,
    codes=[] ) ),
  Element( u'INS14', Properties(desc=u'City Name', req_sit=u'N', data_type=(u'AN',u'2',u'30'), position=14,
    codes=[] ) ),
  Element( u'INS15', Properties(desc=u'State or Province Code', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=15,
    codes=[] ) ),
  Element( u'INS16', Properties(desc=u'Country Code', req_sit=u'N', data_type=(u'ID',u'2',u'3'), position=16,
    codes=[] ) ),
  Element( u'INS17', Properties(desc=u'Number', req_sit=u'R', data_type=(u'N0',u'1',u'9'), position=17,
    codes=[] ) ),
),
Segment( u'DTP', Properties(syntax='',req_sit=u'S',repeat=u'2',pos=u'120',desc=u'Subscriber Date'),
  Element( u'DTP01', Properties(desc=u'Date/Time Qualifier', req_sit=u'R', data_type=(u'ID',u'3',u'3'), position=1,
    codes=[u'102', u'307', u'435', u'472'] ) ),
  Element( u'DTP02', Properties(desc=u'Date Time Period Format Qualifier', req_sit=u'R', data_type=(u'ID',u'2',u'3'), position=2,
    codes=[u'D8', u'RD8'] ) ),
  Element( u'DTP03', Properties(desc=u'Date Time Period', req_sit=u'R', data_type=(u'AN',u'1',u'35'), position=3,
    codes=[] ) ),
),
parsed_270_2110C,
)
parsed_270_2110D = Loop( u'2110D', Properties(looptype='',repeat=u'99',pos=u'130',req_sit=u'R',desc=u'Dependent Eligibility or Benefit Inquiry Information'), 
Segment( u'EQ', Properties(syntax='',req_sit=u'R',repeat=u'1',pos=u'130',desc=u'Dependent Eligibility or Benefit Inquiry Information'),
  Element( u'EQ01', Properties(desc=u'Service Type Code', req_sit=u'S', data_type=(u'ID',u'1',u'2'), position=1,
    codes=[u'1', u'2', u'3', u'4', u'5', u'6', u'7', u'8', u'9', u'10', u'11', u'12', u'13', u'14', u'15', u'16', u'17', u'18', u'19', u'20', u'21', u'22', u'23', u'24', u'25', u'26', u'27', u'28', u'30', u'32', u'33', u'34', u'35', u'36', u'37', u'38', u'39', u'40', u'41', u'42', u'43', u'44', u'45', u'46', u'47', u'48', u'49', u'50', u'51', u'52', u'53', u'54', u'55', u'56', u'57', u'58', u'59', u'60', u'61', u'62', u'63', u'64', u'65', u'66', u'67', u'68', u'69', u'70', u'71', u'72', u'73', u'74', u'75', u'76', u'77', u'78', u'79', u'80', u'81', u'82', u'83', u'84', u'85', u'86', u'87', u'88', u'89', u'90', u'91', u'92', u'93', u'94', u'95', u'96', u'97', u'98', u'99', u'A0', u'A1', u'A2', u'A3', u'A4', u'A5', u'A6', u'A7', u'A8', u'A9', u'AA', u'AB', u'AC', u'AD', u'AE', u'AF', u'AG', u'AH', u'AI', u'AJ', u'AK', u'AL', u'AM', u'AN', u'AO', u'AQ', u'AR', u'BA', u'BB', u'BC', u'BD', u'BE', u'BF', u'BG', u'BH', u'BI', u'BJ', u'BK', u'BL', u'BM', u'BN', u'BP', u'BQ', u'BR', u'BS'] ) ),
  Composite( u'C003', Properties(req_sit=u'S',refdes='',seq=u'02',desc=u'Composite Medical Procedure Identifier'),
    Element( u'EQ02-01', Properties(desc=u'Product/Service ID Qualifier', req_sit=u'R', data_type=(u'ID',u'2',u'2'), position=0,
      codes=[u'AD', u'CJ', u'HC', u'ID', u'IV', u'N4', u'ZZ'] ) ),
    Element( u'EQ02-02', Properties(desc=u'Product/Service ID', req_sit=u'R', data_type=(u'AN',u'1',u'48'), position=1,
      codes=[] ) ),
    Element( u'EQ02-03', Properties(desc=u'Procedure Modifier', req_sit=u'S', data_type=(u'AN',u'2',u'2'), position=2,
      codes=[] ) ),
    Element( u'EQ02-04', Properties(desc=u'Procedure Modifier', req_sit=u'S', data_type=(u'AN',u'2',u'2'), position=3,
      codes=[] ) ),
    Element( u'EQ02-05', Properties(desc=u'Procedure Modifier', req_sit=u'S', data_type=(u'AN',u'2',u'2'), position=4,
      codes=[] ) ),
    Element( u'EQ02-06', Properties(desc=u'Procedure Modifier', req_sit=u'S', data_type=(u'AN',u'2',u'2'), position=5,
      codes=[] ) ),
    Element( u'EQ02-07', Properties(desc=u'Description', req_sit=u'N', data_type=(u'AN',u'1',u'80'), position=6,
      codes=[] ) ),
  ),
  Element( u'EQ03', Properties(desc=u'Coverage Level Code', req_sit=u'S', data_type=(u'ID',u'3',u'3'), position=3,
    codes=[u'CHD', u'DEP', u'ECH', u'EMP', u'ESP', u'FAM', u'IND', u'SPC', u'SPO'] ) ),
  Element( u'EQ04', Properties(desc=u'Insurance Type Code', req_sit=u'S', data_type=(u'ID',u'1',u'3'), position=4,
    codes=[u'AP', u'C1', u'CO', u'GP', u'HM', u'IP', u'MA', u'MB', u'MC', u'PR', u'PS', u'SP', u'WC'] ) ),
),
Segment( u'III', Properties(syntax='',req_sit=u'S',repeat=u'10',pos=u'170',desc=u'Dependent Eligibility or Benefit Additional Inquiry Information'),
  Element( u'III01', Properties(desc=u'Code List Qualifier Code', req_sit=u'R', data_type=(u'ID',u'1',u'3'), position=1,
    codes=[u'BF', u'BK', u'ZZ'] ) ),
  Element( u'III02', Properties(desc=u'Industry Code', req_sit=u'R', data_type=(u'AN',u'1',u'30'), position=2,
    codes=[] ) ),
  Element( u'III03', Properties(desc=u'Code Category', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=3,
    codes=[] ) ),
  Element( u'III04', Properties(desc=u'Free-form Message Text', req_sit=u'N', data_type=(u'AN',u'1',u'264'), position=4,
    codes=[] ) ),
  Element( u'III05', Properties(desc=u'Quantity', req_sit=u'N', data_type=(u'R',u'1',u'15'), position=5,
    codes=[] ) ),
  Composite( u'C001', Properties(req_sit=u'N',refdes='',seq=u'06',desc=u'Composite Unit of Measure'),
  ),
  Element( u'III07', Properties(desc=u'Surface/Layer/Position Code', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=7,
    codes=[] ) ),
  Element( u'III08', Properties(desc=u'Surface/Layer/Position Code', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=8,
    codes=[] ) ),
  Element( u'III09', Properties(desc=u'Surface/Layer/Position Code', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=9,
    codes=[] ) ),
),
Segment( u'REF', Properties(syntax='',req_sit=u'S',repeat=u'1',pos=u'190',desc=u'Dependent Additional Information'),
  Element( u'REF01', Properties(desc=u'Reference Identification Qualifier', req_sit=u'R', data_type=(u'ID',u'2',u'3'), position=1,
    codes=[u'9F', u'G1'] ) ),
  Element( u'REF02', Properties(desc=u'Reference Identification', req_sit=u'R', data_type=(u'AN',u'1',u'50'), position=2,
    codes=[] ) ),
  Element( u'REF03', Properties(desc=u'Description', req_sit=u'N', data_type=(u'AN',u'1',u'80'), position=3,
    codes=[] ) ),
  Composite( u'C040', Properties(req_sit=u'N',refdes='',seq=u'04',desc=u'Reference Identifier'),
  ),
),
Segment( u'DTP', Properties(syntax='',req_sit=u'S',repeat=u'1',pos=u'200',desc=u'Dependent Eligibility/Benefit Date'),
  Element( u'DTP01', Properties(desc=u'Date/Time Qualifier', req_sit=u'R', data_type=(u'ID',u'3',u'3'), position=1,
    codes=[u'307', u'435', u'472'] ) ),
  Element( u'DTP02', Properties(desc=u'Date Time Period Format Qualifier', req_sit=u'R', data_type=(u'ID',u'2',u'3'), position=2,
    codes=[u'D8', u'RD8'] ) ),
  Element( u'DTP03', Properties(desc=u'Date Time Period', req_sit=u'R', data_type=(u'AN',u'1',u'35'), position=3,
    codes=[] ) ),
),
)
parsed_270_2100D = Loop( u'2100D', Properties(looptype='',repeat=u'1',pos=u'030',req_sit=u'R',desc=u'Dependent Name'), 
Segment( u'NM1', Properties(syntax='',req_sit=u'R',repeat=u'1',pos=u'030',desc=u'Dependent Name'),
  Element( u'NM101', Properties(desc=u'Entity Identifier Code', req_sit=u'R', data_type=(u'ID',u'2',u'3'), position=1,
    codes=[u'03'] ) ),
  Element( u'NM102', Properties(desc=u'Entity Type Qualifier', req_sit=u'R', data_type=(u'ID',u'1',u'1'), position=2,
    codes=[u'1'] ) ),
  Element( u'NM103', Properties(desc=u'Name Last or Organization Name', req_sit=u'S', data_type=(u'AN',u'1',u'60'), position=3,
    codes=[] ) ),
  Element( u'NM104', Properties(desc=u'Name First', req_sit=u'S', data_type=(u'AN',u'1',u'35'), position=4,
    codes=[] ) ),
  Element( u'NM105', Properties(desc=u'Name Middle', req_sit=u'S', data_type=(u'AN',u'1',u'25'), position=5,
    codes=[] ) ),
  Element( u'NM106', Properties(desc=u'Name Prefix', req_sit=u'N', data_type=(u'AN',u'1',u'10'), position=6,
    codes=[] ) ),
  Element( u'NM107', Properties(desc=u'Name Suffix', req_sit=u'S', data_type=(u'AN',u'1',u'10'), position=7,
    codes=[] ) ),
  Element( u'NM108', Properties(desc=u'Identification Code Qualifier', req_sit=u'N', data_type=(u'ID',u'1',u'2'), position=8,
    codes=[] ) ),
  Element( u'NM109', Properties(desc=u'Identification Code', req_sit=u'N', data_type=(u'AN',u'2',u'80'), position=9,
    codes=[] ) ),
  Element( u'NM110', Properties(desc=u'Entity Relationship Code', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=10,
    codes=[] ) ),
  Element( u'NM111', Properties(desc=u'Entity Identifier Code', req_sit=u'N', data_type=(u'ID',u'2',u'3'), position=11,
    codes=[] ) ),
),
Segment( u'REF', Properties(syntax='',req_sit=u'S',repeat=u'9',pos=u'040',desc=u'Dependent Additional Identification'),
  Element( u'REF01', Properties(desc=u'Reference Identification Qualifier', req_sit=u'R', data_type=(u'ID',u'2',u'3'), position=1,
    codes=[u'18', u'1L', u'6P', u'A6', u'CT', u'EA', u'EJ', u'F6', u'GH', u'HJ', u'IF', u'IG', u'N6', u'SY'] ) ),
  Element( u'REF02', Properties(desc=u'Reference Identification', req_sit=u'R', data_type=(u'AN',u'1',u'50'), position=2,
    codes=[] ) ),
  Element( u'REF03', Properties(desc=u'Description', req_sit=u'N', data_type=(u'AN',u'1',u'80'), position=3,
    codes=[] ) ),
  Composite( u'C040', Properties(req_sit=u'N',refdes='',seq=u'04',desc=u'Reference Identifier'),
  ),
),
Segment( u'N3', Properties(syntax='',req_sit=u'S',repeat=u'1',pos=u'060',desc=u'Dependent Address'),
  Element( u'N301', Properties(desc=u'Address Information', req_sit=u'R', data_type=(u'AN',u'1',u'55'), position=1,
    codes=[] ) ),
  Element( u'N302', Properties(desc=u'Address Information', req_sit=u'S', data_type=(u'AN',u'1',u'55'), position=2,
    codes=[] ) ),
),
Segment( u'N4', Properties(syntax='',req_sit=u'S',repeat=u'1',pos=u'070',desc=u'Dependent City/State/ZIP Code'),
  Element( u'N401', Properties(desc=u'City Name', req_sit=u'S', data_type=(u'AN',u'2',u'30'), position=1,
    codes=[] ) ),
  Element( u'N402', Properties(desc=u'State or Province Code', req_sit=u'S', data_type=(u'ID',u'2',u'2'), position=2,
    codes=[] ) ),
  Element( u'N403', Properties(desc=u'Postal Code', req_sit=u'S', data_type=(u'ID',u'3',u'15'), position=3,
    codes=[] ) ),
  Element( u'N404', Properties(desc=u'Country Code', req_sit=u'S', data_type=(u'ID',u'2',u'3'), position=4,
    codes=[] ) ),
  Element( u'N405', Properties(desc=u'Location Qualifier', req_sit=u'N', data_type=(u'ID',u'1',u'2'), position=5,
    codes=[] ) ),
  Element( u'N406', Properties(desc=u'Location Identifier', req_sit=u'N', data_type=(u'AN',u'1',u'30'), position=6,
    codes=[] ) ),
),
Segment( u'PRV', Properties(syntax='',req_sit=u'S',repeat=u'1',pos=u'090',desc=u'Provider Information'),
  Element( u'PRV01', Properties(desc=u'Provider Code', req_sit=u'R', data_type=(u'ID',u'1',u'3'), position=1,
    codes=[u'AD', u'AT', u'BI', u'CO', u'CV', u'H', u'HH', u'LA', u'OT', u'P1', u'P2', u'PC', u'PE', u'R', u'RF', u'SB', u'SK', u'SU'] ) ),
  Element( u'PRV02', Properties(desc=u'Reference Identification Qualifier', req_sit=u'R', data_type=(u'ID',u'2',u'3'), position=2,
    codes=[u'9K', u'D3', u'EI', u'HPI', u'SY', u'TJ', u'ZZ'] ) ),
  Element( u'PRV03', Properties(desc=u'Reference Identification', req_sit=u'R', data_type=(u'AN',u'1',u'50'), position=3,
    codes=[] ) ),
  Element( u'PRV04', Properties(desc=u'State or Province Code', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=4,
    codes=[] ) ),
  Composite( u'C035', Properties(req_sit=u'N',refdes='',seq=u'05',desc=u'Provider Specialty Information'),
  ),
  Element( u'PRV06', Properties(desc=u'Provider Organization Code', req_sit=u'N', data_type=(u'ID',u'3',u'3'), position=6,
    codes=[] ) ),
),
Segment( u'DMG', Properties(syntax='',req_sit=u'S',repeat=u'1',pos=u'100',desc=u'Dependent Demographic Information'),
  Element( u'DMG01', Properties(desc=u'Date Time Period Format Qualifier', req_sit=u'S', data_type=(u'ID',u'2',u'3'), position=1,
    codes=[u'D8'] ) ),
  Element( u'DMG02', Properties(desc=u'Date Time Period', req_sit=u'S', data_type=(u'AN',u'1',u'35'), position=2,
    codes=[] ) ),
  Element( u'DMG03', Properties(desc=u'Gender Code', req_sit=u'S', data_type=(u'ID',u'1',u'1'), position=3,
    codes=[u'F', u'M'] ) ),
  Element( u'DMG04', Properties(desc=u'Marital Status Code', req_sit=u'N', data_type=(u'ID',u'1',u'1'), position=4,
    codes=[] ) ),
  Element( u'DMG05', Properties(desc=u'Race or Ethnicity Code', req_sit=u'N', data_type=(u'ID',u'1',u'1'), position=5,
    codes=[] ) ),
  Element( u'DMG06', Properties(desc=u'Citizenship Status Code', req_sit=u'N', data_type=(u'ID',u'1',u'2'), position=6,
    codes=[] ) ),
  Element( u'DMG07', Properties(desc=u'Country Code', req_sit=u'N', data_type=(u'ID',u'2',u'3'), position=7,
    codes=[] ) ),
  Element( u'DMG08', Properties(desc=u'Basis of Verification Code', req_sit=u'N', data_type=(u'ID',u'1',u'2'), position=8,
    codes=[] ) ),
  Element( u'DMG09', Properties(desc=u'Quantity', req_sit=u'N', data_type=(u'R',u'1',u'15'), position=9,
    codes=[] ) ),
),
Segment( u'INS', Properties(syntax='',req_sit=u'S',repeat=u'1',pos=u'110',desc=u'Dependent Relationship'),
  Element( u'INS01', Properties(desc=u'Yes/No Condition or Response Code', req_sit=u'R', data_type=(u'ID',u'1',u'1'), position=1,
    codes=[u'N'] ) ),
  Element( u'INS02', Properties(desc=u'Individual Relationship Code', req_sit=u'R', data_type=(u'ID',u'2',u'2'), position=2,
    codes=[u'01', u'19', u'34'] ) ),
  Element( u'INS03', Properties(desc=u'Maintenance Type Code', req_sit=u'N', data_type=(u'ID',u'3',u'3'), position=3,
    codes=[] ) ),
  Element( u'INS04', Properties(desc=u'Maintenance Reason Code', req_sit=u'N', data_type=(u'ID',u'2',u'3'), position=4,
    codes=[] ) ),
  Element( u'INS05', Properties(desc=u'Benefit Status Code', req_sit=u'N', data_type=(u'ID',u'1',u'1'), position=5,
    codes=[] ) ),
  Element( u'INS06', Properties(desc=u'Medicare Plan Code', req_sit=u'N', data_type=(u'ID',u'1',u'1'), position=6,
    codes=[] ) ),
  Element( u'INS07', Properties(desc=u'Consolidated Omnibus Budget Reconciliation Act (COBRA) Qualifying', req_sit=u'N', data_type=(u'ID',u'1',u'2'), position=7,
    codes=[] ) ),
  Element( u'INS08', Properties(desc=u'Employment Status Code', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=8,
    codes=[] ) ),
  Element( u'INS09', Properties(desc=u'Student Status Code', req_sit=u'N', data_type=(u'ID',u'1',u'1'), position=9,
    codes=[] ) ),
  Element( u'INS10', Properties(desc=u'Yes/No Condition or Response Code', req_sit=u'N', data_type=(u'ID',u'1',u'1'), position=10,
    codes=[] ) ),
  Element( u'INS11', Properties(desc=u'Date Time Period Format Qualifier', req_sit=u'N', data_type=(u'ID',u'2',u'3'), position=11,
    codes=[] ) ),
  Element( u'INS12', Properties(desc=u'Date Time Period', req_sit=u'N', data_type=(u'AN',u'1',u'35'), position=12,
    codes=[] ) ),
  Element( u'INS13', Properties(desc=u'Confidentiality Code', req_sit=u'N', data_type=(u'ID',u'1',u'1'), position=13,
    codes=[] ) ),
  Element( u'INS14', Properties(desc=u'City Name', req_sit=u'N', data_type=(u'AN',u'2',u'30'), position=14,
    codes=[] ) ),
  Element( u'INS15', Properties(desc=u'State or Province Code', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=15,
    codes=[] ) ),
  Element( u'INS16', Properties(desc=u'Country Code', req_sit=u'N', data_type=(u'ID',u'2',u'3'), position=16,
    codes=[] ) ),
  Element( u'INS17', Properties(desc=u'Number', req_sit=u'S', data_type=(u'N0',u'1',u'9'), position=17,
    codes=[] ) ),
),
Segment( u'DTP', Properties(syntax='',req_sit=u'S',repeat=u'2',pos=u'120',desc=u'Dependent Date'),
  Element( u'DTP01', Properties(desc=u'Date/Time Qualifier', req_sit=u'R', data_type=(u'ID',u'3',u'3'), position=1,
    codes=[u'102', u'307', u'435', u'472'] ) ),
  Element( u'DTP02', Properties(desc=u'Date Time Period Format Qualifier', req_sit=u'R', data_type=(u'ID',u'2',u'3'), position=2,
    codes=[u'D8', u'RD8'] ) ),
  Element( u'DTP03', Properties(desc=u'Date Time Period', req_sit=u'R', data_type=(u'AN',u'1',u'35'), position=3,
    codes=[] ) ),
),
parsed_270_2110D,
)
parsed_270_2000D = Loop( u'2000D', Properties(looptype='',repeat=u'>1',pos=u'040',req_sit=u'S',desc=u'Dependent Level'), 
Segment( u'HL', Properties(syntax='',req_sit=u'R',repeat=u'1',pos=u'010',desc=u'Dependent Level'),
  Element( u'HL01', Properties(desc=u'Hierarchical ID Number', req_sit=u'R', data_type=(u'AN',u'1',u'12'), position=1,
    codes=[] ) ),
  Element( u'HL02', Properties(desc=u'Hierarchical Parent ID Number', req_sit=u'R', data_type=(u'AN',u'1',u'12'), position=2,
    codes=[] ) ),
  Element( u'HL03', Properties(desc=u'Hierarchical Level Code', req_sit=u'R', data_type=(u'ID',u'1',u'2'), position=3,
    codes=[u'23'] ) ),
  Element( u'HL04', Properties(desc=u'Hierarchical Child Code', req_sit=u'R', data_type=(u'ID',u'1',u'1'), position=4,
    codes=[u'0'] ) ),
),
Segment( u'TRN', Properties(syntax='',req_sit=u'S',repeat=u'2',pos=u'020',desc=u'Dependent Trace Number'),
  Element( u'TRN01', Properties(desc=u'Trace Type Code', req_sit=u'R', data_type=(u'ID',u'1',u'2'), position=1,
    codes=[u'1'] ) ),
  Element( u'TRN02', Properties(desc=u'Reference Identification', req_sit=u'R', data_type=(u'AN',u'1',u'50'), position=2,
    codes=[] ) ),
  Element( u'TRN03', Properties(desc=u'Originating Company Identifier', req_sit=u'R', data_type=(u'AN',u'10',u'10'), position=3,
    codes=[] ) ),
  Element( u'TRN04', Properties(desc=u'Reference Identification', req_sit=u'S', data_type=(u'AN',u'1',u'50'), position=4,
    codes=[] ) ),
),
parsed_270_2100D,
)
parsed_270_2000C = Loop( u'2000C', Properties(looptype='',repeat=u'>1',pos=u'040',req_sit=u'R',desc=u'Subscriber Level'), 
Segment( u'HL', Properties(syntax='',req_sit=u'R',repeat=u'1',pos=u'010',desc=u'Subscriber Level'),
  Element( u'HL01', Properties(desc=u'Hierarchical ID Number', req_sit=u'R', data_type=(u'AN',u'1',u'12'), position=1,
    codes=[] ) ),
  Element( u'HL02', Properties(desc=u'Hierarchical Parent ID Number', req_sit=u'R', data_type=(u'AN',u'1',u'12'), position=2,
    codes=[] ) ),
  Element( u'HL03', Properties(desc=u'Hierarchical Level Code', req_sit=u'R', data_type=(u'ID',u'1',u'2'), position=3,
    codes=[u'22'] ) ),
  Element( u'HL04', Properties(desc=u'Hierarchical Child Code', req_sit=u'R', data_type=(u'ID',u'1',u'1'), position=4,
    codes=[u'0', u'1'] ) ),
),
Segment( u'TRN', Properties(syntax='',req_sit=u'S',repeat=u'2',pos=u'020',desc=u'Subscriber Trace Number'),
  Element( u'TRN01', Properties(desc=u'Trace Type Code', req_sit=u'R', data_type=(u'ID',u'1',u'2'), position=1,
    codes=[u'1'] ) ),
  Element( u'TRN02', Properties(desc=u'Reference Identification', req_sit=u'R', data_type=(u'AN',u'1',u'50'), position=2,
    codes=[] ) ),
  Element( u'TRN03', Properties(desc=u'Originating Company Identifier', req_sit=u'R', data_type=(u'AN',u'10',u'10'), position=3,
    codes=[] ) ),
  Element( u'TRN04', Properties(desc=u'Reference Identification', req_sit=u'S', data_type=(u'AN',u'1',u'50'), position=4,
    codes=[] ) ),
),
parsed_270_2100C,
parsed_270_2000D,
)
parsed_270_2000B = Loop( u'2000B', Properties(looptype='',repeat=u'>1',pos=u'040',req_sit=u'R',desc=u'Information Receiver Level'), 
Segment( u'HL', Properties(syntax='',req_sit=u'R',repeat=u'1',pos=u'010',desc=u'Information Receiver Level'),
  Element( u'HL01', Properties(desc=u'Hierarchical ID Number', req_sit=u'R', data_type=(u'AN',u'1',u'12'), position=1,
    codes=[] ) ),
  Element( u'HL02', Properties(desc=u'Hierarchical Parent ID Number', req_sit=u'R', data_type=(u'AN',u'1',u'12'), position=2,
    codes=[] ) ),
  Element( u'HL03', Properties(desc=u'Hierarchical Level Code', req_sit=u'R', data_type=(u'ID',u'1',u'2'), position=3,
    codes=[u'21'] ) ),
  Element( u'HL04', Properties(desc=u'Hierarchical Child Code', req_sit=u'R', data_type=(u'ID',u'1',u'1'), position=4,
    codes=[u'1'] ) ),
),
parsed_270_2100B,
parsed_270_2000C,
)
parsed_270_2000A = Loop( u'2000A', Properties(looptype='',repeat=u'>1',pos=u'010',req_sit=u'R',desc=u'Information Source Level'), 
Segment( u'HL', Properties(syntax='',req_sit=u'R',repeat=u'1',pos=u'010',desc=u'Information Source Level'),
  Element( u'HL01', Properties(desc=u'Hierarchical ID Number', req_sit=u'R', data_type=(u'AN',u'1',u'12'), position=1,
    codes=[] ) ),
  Element( u'HL02', Properties(desc=u'Hierarchical Parent ID Number', req_sit=u'N', data_type=(u'AN',u'1',u'12'), position=2,
    codes=[] ) ),
  Element( u'HL03', Properties(desc=u'Hierarchical Level Code', req_sit=u'R', data_type=(u'ID',u'1',u'2'), position=3,
    codes=[u'20'] ) ),
  Element( u'HL04', Properties(desc=u'Hierarchical Child Code', req_sit=u'R', data_type=(u'ID',u'1',u'1'), position=4,
    codes=[u'1'] ) ),
),
parsed_270_2100A,
parsed_270_2000B,
)
parsed_270_DETAIL = Loop( u'DETAIL', Properties(looptype=u'wrapper',repeat=u'>1',pos=u'020',req_sit=u'S',desc=u'Table 2 - Detail'), 
parsed_270_2000A,
)
parsed_270_ST_LOOP = Loop( u'ST_LOOP', Properties(looptype=u'explicit',repeat=u'>1',pos=u'020',req_sit=u'R',desc=u'Transaction Set Header'), 
Segment( u'ST', Properties(syntax='',req_sit=u'R',repeat=u'1',pos=u'010',desc=u'Transaction Set Header'),
  Element( u'ST01', Properties(desc=u'Transaction Set Identifier Code', req_sit=u'R', data_type=(u'ID',u'3',u'3'), position=1,
    codes=[u'270'] ) ),
  Element( u'ST02', Properties(desc=u'Transaction Set Control Number', req_sit=u'R', data_type=(u'AN',u'4',u'9'), position=2,
    codes=[] ) ),
),
parsed_270_HEADER,
parsed_270_DETAIL,
Segment( u'SE', Properties(syntax='',req_sit=u'R',repeat=u'1',pos=u'210',desc=u'Transaction Set Trailer'),
  Element( u'SE01', Properties(desc=u'Number of Included Segments', req_sit=u'R', data_type=(u'N0',u'1',u'10'), position=1,
    codes=[] ) ),
  Element( u'SE02', Properties(desc=u'Transaction Set Control Number', req_sit=u'R', data_type=(u'AN',u'4',u'9'), position=2,
    codes=[] ) ),
),
)
parsed_270_GS_LOOP = Loop( u'GS_LOOP', Properties(looptype=u'explicit',repeat=u'>1',pos=u'020',req_sit=u'R',desc=u'Functional Group Header'), 
Segment( u'GS', Properties(syntax='',req_sit=u'R',repeat=u'1',pos=u'010',desc=u'Functional Group Header'),
  Element( u'GS01', Properties(desc=u'Functional Identifier Code', req_sit=u'R', data_type=(u'ID',u'2',u'2'), position=1,
    codes=[u'HS'] ) ),
  Element( u'GS02', Properties(desc=u'Application Senders Code', req_sit=u'R', data_type=(u'AN',u'2',u'15'), position=2,
    codes=[] ) ),
  Element( u'GS03', Properties(desc=u'124', req_sit=u'R', data_type=(u'AN',u'2',u'15'), position=3,
    codes=[] ) ),
  Element( u'GS04', Properties(desc=u'Date', req_sit=u'R', data_type=(u'DT',u'8',u'8'), position=4,
    codes=[] ) ),
  Element( u'GS05', Properties(desc=u'Time', req_sit=u'R', data_type=(u'TM',u'4',u'8'), position=5,
    codes=[] ) ),
  Element( u'GS06', Properties(desc=u'Group Control Number', req_sit=u'R', data_type=(u'N0',u'1',u'9'), position=6,
    codes=[] ) ),
  Element( u'GS07', Properties(desc=u'Responsible Agency Code', req_sit=u'R', data_type=(u'ID',u'1',u'2'), position=7,
    codes=[u'X'] ) ),
  Element( u'GS08', Properties(desc=u'Version / Release / Industry Identifier Code', req_sit=u'R', data_type=(u'AN',u'1',u'12'), position=8,
    codes=[u'004010X092A1'] ) ),
),
parsed_270_ST_LOOP,
Segment( u'GE', Properties(syntax='',req_sit=u'R',repeat=u'1',pos=u'030',desc=u'Functional Group Trailer'),
  Element( u'GE01', Properties(desc=u'97', req_sit=u'R', data_type=(u'N0',u'1',u'6'), position=1,
    codes=[] ) ),
  Element( u'GE02', Properties(desc=u'Group Control Number', req_sit=u'R', data_type=(u'N0',u'1',u'9'), position=2,
    codes=[] ) ),
),
)
parsed_270_ISA_LOOP = Loop( u'ISA_LOOP', Properties(looptype=u'explicit',repeat=u'>1',pos=u'001',req_sit=u'R',desc=u'Interchange Control Header'), 
Segment( u'ISA', Properties(syntax='',req_sit=u'R',repeat=u'1',pos=u'010',desc=u'Interchange Control Header'),
  Element( u'ISA01', Properties(desc=u'I01', req_sit=u'R', data_type=(u'ID',u'2',u'2'), position=1,
    codes=[u'00', u'03'] ) ),
  Element( u'ISA02', Properties(desc=u'I02', req_sit=u'R', data_type=(u'AN',u'10',u'10'), position=2,
    codes=[] ) ),
  Element( u'ISA03', Properties(desc=u'I03', req_sit=u'R', data_type=(u'ID',u'2',u'2'), position=3,
    codes=[u'00', u'01'] ) ),
  Element( u'ISA04', Properties(desc=u'I04', req_sit=u'R', data_type=(u'AN',u'10',u'10'), position=4,
    codes=[] ) ),
  Element( u'ISA05', Properties(desc=u'I05', req_sit=u'R', data_type=(u'ID',u'2',u'2'), position=5,
    codes=[u'01', u'14', u'20', u'27', u'28', u'29', u'30', u'33', u'ZZ'] ) ),
  Element( u'ISA06', Properties(desc=u'I06', req_sit=u'R', data_type=(u'AN',u'15',u'15'), position=6,
    codes=[] ) ),
  Element( u'ISA07', Properties(desc=u'I05', req_sit=u'R', data_type=(u'ID',u'2',u'2'), position=7,
    codes=[u'01', u'14', u'20', u'27', u'28', u'29', u'30', u'33', u'ZZ'] ) ),
  Element( u'ISA08', Properties(desc=u'I07', req_sit=u'R', data_type=(u'AN',u'15',u'15'), position=8,
    codes=[] ) ),
  Element( u'ISA09', Properties(desc=u'I08', req_sit=u'R', data_type=(u'DT',u'6',u'6'), position=9,
    codes=[] ) ),
  Element( u'ISA10', Properties(desc=u'I09', req_sit=u'R', data_type=(u'TM',u'4',u'4'), position=10,
    codes=[] ) ),
  Element( u'ISA11', Properties(desc=u'I10', req_sit=u'R', data_type=(u'ID',u'1',u'1'), position=11,
    codes=[u'U'] ) ),
  Element( u'ISA12', Properties(desc=u'I11', req_sit=u'R', data_type=(u'ID',u'5',u'5'), position=12,
    codes=[u'00401'] ) ),
  Element( u'ISA13', Properties(desc=u'I12', req_sit=u'R', data_type=(u'N0',u'9',u'9'), position=13,
    codes=[] ) ),
  Element( u'ISA14', Properties(desc=u'I13', req_sit=u'R', data_type=(u'ID',u'1',u'1'), position=14,
    codes=[u'0', u'1'] ) ),
  Element( u'ISA15', Properties(desc=u'I14', req_sit=u'R', data_type=(u'ID',u'1',u'1'), position=15,
    codes=[u'P', u'T'] ) ),
  Element( u'ISA16', Properties(desc=u'I15', req_sit=u'R', data_type=(u'AN',u'1',u'1'), position=16,
    codes=[] ) ),
),
parsed_270_GS_LOOP,
Segment( u'TA1', Properties(syntax='',req_sit=u'S',repeat=u'1',pos=u'020',desc=u'Interchange Acknowledgement'),
  Element( u'TA101', Properties(desc=u'I12', req_sit=u'R', data_type=(u'N0',u'9',u'9'), position=1,
    codes=[] ) ),
  Element( u'TA102', Properties(desc=u'I08', req_sit=u'R', data_type=(u'DT',u'6',u'6'), position=2,
    codes=[] ) ),
  Element( u'TA103', Properties(desc=u'I09', req_sit=u'R', data_type=(u'TM',u'4',u'4'), position=3,
    codes=[] ) ),
  Element( u'TA104', Properties(desc=u'I17', req_sit=u'R', data_type=(u'ID',u'1',u'1'), position=4,
    codes=[u'A', u'E', u'R'] ) ),
  Element( u'TA105', Properties(desc=u'I18', req_sit=u'R', data_type=(u'ID',u'3',u'3'), position=5,
    codes=[u'000', u'001', u'002', u'003', u'004', u'005', u'006', u'007', u'008', u'009', u'010', u'011', u'012', u'013', u'014', u'015', u'016', u'017', u'018', u'019', u'020', u'021', u'022', u'023', u'024', u'025', u'026', u'027', u'028', u'029', u'030', u'031'] ) ),
),
Segment( u'IEA', Properties(syntax='',req_sit=u'R',repeat=u'1',pos=u'030',desc=u'Interchange Control Trailer'),
  Element( u'IEA01', Properties(desc=u'I16', req_sit=u'R', data_type=(u'N0',u'1',u'5'), position=1,
    codes=[] ) ),
  Element( u'IEA02', Properties(desc=u'I12', req_sit=u'R', data_type=(u'N0',u'9',u'9'), position=2,
    codes=[] ) ),
),
)
parsed_270 = Message( u'270', Properties(desc=u'HIPAA Health Care Eligibility Inquiry X092A1-270'), 
parsed_270_ISA_LOOP,
)




"""
Facade Use Cases
================

The :class:`X12LoopBridge` will associate a Python class with an X12 Loop definition.
Each attribute of Python class is a **Descriptor** that references an X12 Element
in  a Segment instance within the overall Loop.

It's a **Bridge** between a Python class and an X12 Loop.

Repeating copies of a Loop create multiple object instances.

Repeating copies of a Segment within a Loop lead to a
somewhat more complex definition.  These segments are "unwrapped" by the
containing Loop.

A Python **Descriptor** gives us an implicit get/set/del protocol around the
attribute. Syntactically, an application simply references the attribute name.
Python will use the get and set methods from the Descriptor to access the
attribute.

The attribute Descriptor has several use cases, depending
on the attribute's type and how that value is accessed from the X12 Segment.

    -   Single simple value, identified by position within a Segment of the top-level Loop.
        This is a direct reference to an Element, and is only appropriate when the
        Segment doesn't repeat.

    -   Single simple value, one of a number of alternative positions within a Segment.
        This is a
        direct reference to an Element, and is appropriate when the Segment
        doesn't repeat.  The Element is one of a number of
        candidates, which are defined by a sequence of
        :samp:`(qualifier,value)` position tuples.  If the Element value in the qualifier
        position is equal to a given qualifier value, the Element in the value
        position is the desired Element. This is required to pick values out of
        the :samp:`PER` segment.

    -   Sequence of simple values, with the value identified a fixed position in multiple
        occurrences of a Segment.  This is used for :samp:`NTE` segments.

    -   Single simple value, identified by a position in a Composite.  This is a direct reference
        to a sub-element of a Composite.  This is used to pick apart the various
        composites that occur in :samp:`HI` segments when a single value is required.

    -   Single complex value, identified by a Composite.

    -   Sequence of simple values, with the value identified by a fixed position in
        multiple occurrences of a Composite.  This is used to pick apart the various
        composites that occur in :samp:`HI` segments when a sequence of values is required.

    -   Sequence of complex values, with the value identified by a recurring
        Segment.  This is used to define a nested object.  This is used to pick apart the :samp:`PWK`
        Segments, which occur multiple times within the 2300 and 2400 loops.

XXX - Can Composite and Element be refactored to have a common superclass?
Do all Composites have a type identifier in the initial position?

There are several dimensions to the above cases:

    -   Source Type:  Segment (which has an underlying Segment Token) and
        Composite (which is just a string).  This distinction applies to
        some of the following dimensions.

    -   Source Objects:

        -   Segment or Segments that match simple qualifications.

        -   Composite or Composites that match simple Criteria, but are found in multiple Segments.

    -   Element Selection:

        -   Fixed position within each Segment, OneOf a Number of Positions
            within a Segment.  This creates a sequence of strings values.

        -   Fixed position(s) within a Composite, Sequence of Positions
            within a Composite (usually positions [1:]).  This creates a sequence
            of string values.

        -   The whole Segment.

    -   Conversion: No conversion, single-element conversion, multi-element conversion.
        At this point, we don't care whether the data was from a Composite or Segment
        since it's now a sequence of strings.

    -   Result Stucture: single occurance vs. the complete sequence.

There are several conversion protocols.

    -   Single-element string value is the native X12 source data types.
        No conversion is done.

    -   Single-element non-string value is done through a simple conversion
        protocol.  A class with two static methods (x12_to_python and python_to_x12)
        is used by the Descriptor to make the value usable.

    -   Multiple-element all-string values are can be handled as simple Python lists of
        strings.

    -   Multiple-element, non-String values use the same basic conversion
        protocol, starting from a list of strings instead of a single string Element
        value.

Facade Implementation
============================

The :class:`X12LoopBridge` handles the high-level mapping between a Python class
and an instance of an :class:`X12.message.X12Loop`. Each bridge has a few
high-level methods, and a long sequence of attribute definitions. Each attribute
is defined using an instance of the :class:`ElementAccess` class.

The :class:`ElementAccess` class is a **Lightweight** class that has parameters
and a reference to the underlying :class:`X12.message.X12Loop` structure. It is
a Python **Descriptor**, and implements the get/set protocol for the attribute
value.

..  autoclass:: X12LoopBridge

..  autoclass:: X12SegmentBridge

Element Access
--------------

The essential feature of Element Access is to provide a two-way mapping between
a named attribute and Elements within a Segment.  This is done through two
parallel multi-step alogithms for locating the element and either getting
or setting the value.

The element access get algorithm has the following steps:

    1.  **Source Segments**. Accumulate the source X12Segment instances from
        the Loop. Filtering is a feature of the :class:`X12LoopBridge`.

    2.  If doing whole Segments:

        -   Repackage the Segment as a list of strings for step 5.

    3.  If doing Elements from Segments (via :class:`ElementAccess`):

        -   **Element Selection**. Locate actual positions of Element values
            within each Segment. This uses a sequence of X12Segment instances.
            There are several
            segment Position algorithms: simple value position, qualified
            position.  Package as list of strings for step 5.

    4.  Elif doing Elements from Composites (via :class:`CompositeAccess`):

        -   **Source Composites**. Accumulate the source Composite sequences.
            Composite filtering is a feature an {CompositeAccess} object, it
            leverages the source segment list.

        -   **Element Selection**. Accumulate the sequence of Element values.
            This is done by a :class:`CompositePosition` object.  There are two
            composite positions algorithms: single value position and range of positions.
            Package as a list of strings for step 5.

    5.  **Conversion**. Apply x12_to_python type conversion to sequences of Element values.
        Ths various :class:`Conversion` subclasses handle this.

    6.  **Result Structure**. Pick the first value or the whole list of values.
        The :class:`ElementAccess` class handles this final selection.

The set algorithm is the approximate inverse of this.

    1.  **Conversion**.  Apply python_to_x12 type conversion to build sequences
        of Element values.

    2.  Then the **Source Objects** and **Element Selection** steps are done to
        locate through the applicable source element(s).

..  autoclass:: SegmentSequenceAccess

..  autoclass:: ElementAccess

..  autoclass:: ElementSequenceAccess

Segment Position
-----------------

Element values are located by "position".  In the majority of cases, the
position is a single, fixed position within a Segment.  The position can
be defined numerically.

When creating an :class:`ElementAccess` attribute definition, a numeric position
is implicitly translated to an instance of :class:`Position`, the simplest kind
of position definition.

There are several kinds of Position definitions.

    -   :class:`Position` defines a single, fixed Element position as the value
        for an attribute.

    -   :class:`OneOf` is used to implement attributes of the :samp:`PER` Segment, which
        are defined by a series of (qualifer,value) pairs.  The Element's position
        isn't fixed, but is determined by examining a number of alterantives.

    -   :class:`SequenceOf` is used to implement attributes which are a simple sequence
        of values picked from the Elements of the Segment.

..  autoclass:: Position

..  autoclass:: OneOf

..  autoclass:: SequenceOf

Composite Position
------------------

Element values are located by "position".  Positions are either a single,
fixed position within a Composite or a range of positions.

..  autoclass:: CompositePosition

..  autoclass:: CompositeAccess

# XXX - Refactor CompositeAccess and CompositeSequenceAccess to be much simpler.

..  autoclass:: CompositeSequenceAccess

Conversion
----------

Each data conversion is a static class definition with two methods:

    -   x12_to_python.  This method converts an X12 string or sequence
        of strings into a Python object.  The sequence of strings is used
        to convert Composites as well as converting whole Segments into distinct
        Python objects.

    -   python_to_x12.  This method converts a Python object into an
        X12 string or a sequence of strings.

..  autoclass:: Conversion
..  autoclass:: D8
..  autoclass:: DR
..  autoclass:: SegmentConversion


Exceptions
-------------

..  autoclass:: MissingSegment
"""
import datetime
from decimal import Decimal


class Facade(object):
    def loops(self, theClass, anX12Message, *args, **kwargs):
        return [theClass(loop, *args, **kwargs) for loop in
                anX12Message.descendant("loop", theClass.loopName)]


class MissingSegment( Exception ):
    """This exception is raised if the target Segment cannot be found
    within the Loop.
    """
    pass

class X12LoopBridge( object ):
    """Bridge between a user-focused class/attribute definition and an :class:`X12.message.X12Loop`.

    This wrapper implements a simple search for Segments within a Loop.
    It leverates the essential XPath like "descendant" search capability,
    and either returns a sequence of Segments or a single Segment which matches
    certain qualifying criteria.

    This is an abstract superclass definition.  A subclass will contain a number
    of attributes, defined using the :class:`ElementAccess` class or any of its
    subclasses.
    """
    def __init__( self, aLoop ):
        """Named loop within a specific message.
        
        :param aLoop: the loop identifier to pick out of the X12Message.
        """
        self.loop= aLoop
    def __str__( self ):
        return str(self.loop)
    def _filteredList( self, name, qualifierPos=None, inList=None, notInList=None ):
        """Return all matching X12Segments that are children of the X12Loop.
        Do not check descendants, only check immediate children.
        If no qualifier, then all segments with the given name are returned.
        If a qualifier position is provided, then the element at that position in the
        segment is checked for a value in the :meth:`inList` or not in the :meth:`notInList`
        values.
        
        :param name: name of the segment
        :param qualifierPos: optional qualifier position in the Segment.
        if omitted, all matching segments are returned.
        :param inList: optional qualification values.  Only used if the qualiferPosition
        is provided.  This the positive list of values segments must have.
        :param notInList: optional qualification values.  Only used if the qualiferPosition
        is provided.  This the negatuve list of values segments must not have.
        :returns: list of all :class:`X12.message.X12Segment` instances within this Loop.
        """
        segList= [ s for s in self.loop.child( "segment", name ) ]
        if qualifierPos:
            if inList:
                filtered= [ seg for seg in segList if seg.getByPos(qualifierPos) in inList ]
            elif notInList:
                filtered= [ seg for seg in segList if seg.getByPos(qualifierPos) not in notInList ]
            else:
                raise TypeError( "QualifierPos requires inList or notInList")
        else:
            filtered= segList
        return filtered
    def segList( self, name, qualifierPos=None, inList=None, notInList=None ):
        """Return a all matching X12Segments of the X12Loop.
        If no qualifier, then all segments with the given name are returned.
        If a qualifier position is provided, then the element at that position in the
        segment is checked for a value in the :meth:`inList` or not in the :meth:`notInList`
        values.
        
        :param name: name of the segment
        :param qualifierPos: optional qualifier position in the Segment.
        if omitted, all matching segments are returned.
        :param inList: optional qualification values.  Only used if the qualiferPosition
        is provided.  This the positive list of values segments must have.
        :param notInList: optional qualification values.  Only used if the qualiferPosition
        is provided.  This the negatuve list of values segments must not have.
        :returns: list of all :class:`X12.message.X12Segment` instances within this Loop.
        """
        return [ X12SegmentBridge(s) for s in self._filteredList(name,qualifierPos,inList,notInList) ]
    def segment( self, name, qualifierPosition=None, inList=None, notInList=None ):
        """Return a specific X12Segment of the X12Loop.
        If no qualifier, then the first segment with the given name is returned.
        If there's one instance of a segment within a loop, this default
        picks the only instance.
        If a qualifier position is provided, then the element at that position in the
        segment is checked for a value in the list of values.
        
        :param name: name of the segment
        :param qualifierPosition: optional qualification position.
        if omitted, the first matching segment is returned.
        :param inList: optional qualification values.  Only used if the qualifier
        is provided to specify the position to check.
        :param notInList: optional negative list of values to exclude from qualification.
        Only used if the qualifier
        is provided to specify the position to check.
        :returns: :class:`X12.message.X12Segment` instance within this Loop.
        """
        matches= self._filteredList(name,qualifierPosition,inList,notInList)
        if len(matches) == 0: return None
        return X12SegmentBridge( matches[0] )

class X12SegmentBridge( object ):
    """Bridge between a user-focused class/attribute definition and an :class:`X12.message.X12Segment`.

    This is a simple wrapper around an X12Segment which can be used to locate Composites
    within the Segment or locate Elements within the Segment.

    XXX - This is only really necessary when extracting Composites.
    Consider not using X12SegmentBridge for ordinary ElementAccess processing.
    """
    def __init__(self, aSegment ):
        self.segment= aSegment
    def __str__( self ):
        return str( self.segment )
    def compositeList( self, *names ):
        sep= self.segment.message.getCompositeSeparator()
        result= []
        pos= 1
        while self.segment.getByPos(pos) is not "":
            composite= self.segment.getByPos( pos )
            # XXX - Seems vaguely icky to parse composites here.
            subElts= composite.split( sep )
            if subElts[0] in names:
                result.append( subElts )
            pos += 1
        return result
    def composite( self, *name ):
        compList= self.compositeList( *name )
        if len(compList) > 0: return compList[0]
        return None # technically redundant to do this.


class SegmentAccess( object ):
    """Used to get a single segment.

    Best used in Conjunction with X12SegmentBridge, eg:
        class NamedEntity(X12SegmentBridge):
            last_name = ElementAccess("NM1", 3)

        class InformationSource(X12LoopBridge):
            loopName = "1000"
            entity_details = SegmentAccess("NM1",
                x12type=SegmentConversion(NamedEntity))

        class InformationReceiver(X12LoopBridge):
            loopName = "2000"
            entity_details = SegmentAccess("NM1",
                x12type=SegmentConversion(NamedEntity))

    """
    def __init__( self, segment, qualifier=None, x12type=None ):
        self.segment= segment
        if qualifier is None:
            self.qualifier= None
        elif isinstance(qualifier, (list,tuple)):
            self.qualifier= qualifier
        else:
            self.qualifier= ( qualifier, )
        self.x12type= x12type

    def __repr__( self ):
        """Provide Documentation for epydoc."""
        typeName = "None" if self.x12type is None else self.x12type.__name__
        return "SegmentAccess( %r, %r, %s )" % ( self.segment, self.qualifier, typeName )

    def __get__( self, instance, owner ):
        """Get the requested Segment and convert it, if applicable.

        :param instance: An X12LoopBridge object.
        """
        if self.qualifier is None:
            segBridge = instance.segment( self.segment, )
        else:
            segBridge = instance.segment( self.segment, self.qualifier[0], inList=self.qualifier[1:] )
        if segBridge is None:
            return None
        return self.x12type.x12_to_python(segBridge.segment)

    def __set__( self, instance, value ):
        raise UnimplementedError( "Can't set segment sequences, yet")


class SegmentSequenceAccess( object ):
    """Define access to sequence of Segments with a user-friendly attribute name.
    This appears as a sequence of individual object instances.
    This requires a :class:`Conversion` for doing data
    type conversion of the Segment's collection of Elements.

    This class is used to create attribute definitions which are
    mappings to X12Segments in an :class:`X12.message.X12Loop` instance.

    This is a Descriptor, which implements the attribute name via implicit
    getters and setters defined in this class.  These are attached to X12LoopBridge
    objects, so "instance" is always a specific X12LoopBridge, and owner is
    a subclass of X12LoopBridge.

        -   __get__ must locate the Segments and provide lists of string element values
            to a Conversion class.

        -   __set__ must convert objects to lists of strings, then locate the elements and replace their
            values.
    """
    def __init__( self, segment, qualifier=None, x12type=None ):
        """Define an attribute.
        Most attributes are the string values of Elements, simply accessed by position.

        If an attribute has a non-string type, the :py:data:`x12type` parameter can identify
        a :class:`Conversion` class to use.

        If an attribute is on a Segment which can occur multiple times, with a qualified
        value, then the :py:data:`qualifier` parameter is used to provide a tuple with
        the position of the segment used for qualification and the value to test.

        For example: :samp:`segment="REF", qualifier=(2,"SY")` will examine all "REF" segments
        within this loop, looking for one with :samp:`"SY"` in :samp:`REF02`.

        :param segment: Name of the segment that contains the element which has the
        value for this attribute.
        :param qualifier: A qualifier tuple used when there are multiple instances
        of the given segment name and a specific position in the segment must be tested.
        :param x12type: An X12type conversion class name.
        :returns: The element value for this attribute name.
        """
        self.segment= segment
        if qualifier is None:
            self.qualifier= None
        elif isinstance(qualifier, (list,tuple)):
            self.qualifier= qualifier
        else:
            self.qualifier= ( qualifier, )
        self.x12type= x12type
    def __repr__( self ):
        """Provide Documentation for epydoc."""
        typeName = "None" if self.x12type is None else self.x12type.__name__
        return "SegmentSequenceAccess( %r, %r, %s )" % ( self.segment, self.qualifier, typeName )
    def __get__( self, instance, owner ):
        # XXX - Some are unfiltered...
        if self.qualifier is None:
            segBridgeList= instance.segList( self.segment, )
        else:
            segBridgeList= instance.segList( self.segment, self.qualifier[0], inList=self.qualifier[1:] )
        return [ self.x12type.x12_to_python(segBridge.segment) for segBridge in segBridgeList ]
    def __set__( self, instance, value ):
        raise UnimplementedError( "Can't set segment sequences, yet")

class Position( object ):
    """Sets or gets the value of an Element in a Segment based on a
    single, fixed position.
    """
    def __init__( self, position ):
        """Define a fixed Position for ElementAccess.
        
        :param position: the numeric position of this element.
        """
        self.position= position
    def __repr__( self ):
        return "Position(%d)" % ( self.position, )
    def get( self, aSegment ):
        """Get our defined Element out of the given X12Segment.
        
        :param aSegment: an :class:`X12.message.X12Segment` to examine.
        """
        return aSegment.getByPos( self.position )
    def set( self, aSegment, value ):
        """Set our defined Element within the given X12Segment.
        
        :param aSegment: an :class:`X12.message.X12Segment` to update.
        :param value: the String value to place into the segment.
        """
        aSegment.setByPos( self.position, value )

class OneOf( Position ):
    """Sets or Gets the value of an Element by searching through a sequence
    of alternative Element pairs using a qualifier value.
    This subclsas of Position is initialized with a sequence (qualifier,value) position pairs and
    a qualifier value.  If the qualifier value matches the element
    in the qualifier position, then the element in the value position
    is used.
    """
    def __init__( self, value, *posPairSeq ):
        """Define a qualifier value and a sequence of (qualifier,value) positions ElementAccess.
        
        :param value: the qualifier value used to match the Element in one of the qualifier positions.
        :param posPairSeq: sequence of (qualifierPosition, valuePosition)
        """
        self.value= value
        self.posPairSeq= posPairSeq
    def __repr__( self ):
        return "OneOf(%r,%r)" % ( self.value, self.posPairSeq, )
    def get( self, aSegment ):
        """Get our defined element out of the given segment.
        If the qualifier value cannot be found, return None.
        
        :param aSegment: an :class:`X12.message.X12Segment` to examine.
        """
        for qualPos, valPos in self.posPairSeq:
            if aSegment.getByPos( qualPos ) == self.value:
                return aSegment.getByPos( valPos )
        return None
    def set( self, aSegment, value ):
        """Set our defined element within the given segment.
        If the qualifier value cannot be found, raise a NotImplementedError.
        
        XXX - Implement the ability to add qualifiers and values to a segment.
        
        :param aSegment: an :class:`X12.message.X12Segment` to examine.
        :param value: the String value to place into the segment.
        """
        for qualPos, valPos in self.posPairSeq:
            if aSegment.getByPos( qualPos ) == self.value:
                aSegment.setByPos( valPos, value )
                return
        raise NotImplementedError( "Unimplemented: add new OneOf value to this segment" )

class SequenceOf( Position ):
    """Unwind a number of Elements into an sequence of attribute values.
    Often, each value is a Composite element with a qualifier, plus additional values."""
    def __init__( self, start, end ):
        """Define a sequence Position for ElementAccess.
        
        :param start: the first Element position within the Segment.
        :param end: last Element position within the Segment)
        """
        self.start= start
        self.end= end
    def __repr__( self ):
        return "SequenceOf(%d,%d)" % ( self.start, self.end, )
    def get( self, aSegment ):
        elementList= [ aSegment.getByPos(pos) for pos in range(self.start,self.end) if aSegment.getByPos(pos) is not None ]
        return elementList
    def set( self, aSegment, valueList ):
        for pos,val in zip(range(self.start,self.end),valueList):
            aSegment.setByPos(pos,val)
            return
        raise NotImplementedError( "Unimplemented: add new SequenceOf value to this segment" )

class ElementAccess( object ):
    """Define access to Element(s) within a Segment with a user-friendly attribute name.
    This can also bind in a :class:`Position` for finding the specific Element, and
    a :class:`Conversion` for doing data type conversion of the Element.

    This class is used to create attribute definitions which are
    mappings to positional Elements in an :class:`X12.message.X12Segment` instance.

    This is a Descriptor, which implements the attribute name via implicit
    getters and setters defined in this class.  These are attached to X12LoopBridge
    objects, so "instance" is always a specific X12LoopBridge, and owner is
    a subclass of X12LoopBridge.

    The instance object that has ElementAccess attribtues may influence the
    default qualifier used. For example:

        class Person(X12LoopBridge):
            first_name = ElementAccess("NM1", 4)
            last_name = ElementAccess("NM1", 3, qualifier=(1, "IL"))

            def __init__(self, aLoop):
                self.qualifier = (1, "QC")
                super(Person, self).__init__(aLoop)

    When you instantiate this Person class, the resulting object will have a
    first_name ElementAccess attributes that uses (1, "QC") as its qualifier,
    and a last_name ElementAccess that uses (1, "IL"). This is pretty awful,
    and I do apologize for the weird behavior but I have a deadline to meet.
    At least I left behind this helpful comment...

        -   __get__ must locate the elements and provide string to a Conversion class.

        -   __set__ must convert strings, then locate the elements and replace their
            values.
    """
    def __init__( self, segment, position=None, oneOf=None, qualifier=None, x12type=None ):
        """Define an attribute.
        Most attributes are the string values of Elements, simply accessed by position.

        If an attribute has a non-string type, the :py:data:`x12type` parameter can identify
        a :class:`Conversion` class to use.

        If an attribute is on a Segment which can occur multiple times, with a qualified
        value, then the :py:data:`qualifier` parameter is used to provide a tuple with
        the position of the segment used for qualification and the value to test.

        For example: :samp:`segment="REF", qualifier=(2,"SY")` will examine all "REF" segments
        within this loop, looking for one with :samp:`"SY"` in :samp:`REF02`.

        If an attribute occurs multiple times within a segment, and a qualifying
        element is used, then the :class:`oneOf` parameter is used.  This accepts a sequence
        of values: the first is a test to use, the remaining values are
        pairs of positions.  The first position is used to test, the second position
        is the result.

        For example: :samp:`segment="PER", oneOf=( "EM", (3,4), (5,6), (7,8) )` means that
        if element 3 is "EM", the attribute is element 4; if element 5 is "EM",
        the attribute is element 6; if element 7 is "EM", the attribute is element 8.

        :param segment: Name of the segment that contains the element which has the
        value for this attribute.
        :param position: integer position of the element within the segment; this
        is not used if oneOf is used to pick a position.
        :param oneOf: a sequence with a qualifier and position pairs used to
        examine a qualifier element and a value element.  Example
        :samp:`oneOf=( "EM", (3,4), (5,6), (7,7)`.  If oneOf is used, position is
        meaningless.
        :param qualifier: A qualifier tuple used when there are multiple instances
        of the given segment name and a specific position in the segment must be tested.
        :param x12type: An X12type conversion class name.
        :returns: The element value for this attribute name.
        """
        self.segment= segment
        if isinstance(position, int):
            self.position= Position( position )
        elif isinstance(oneOf, (list,tuple) ):
            self.position= OneOf( *oneOf )
        elif isinstance( position, Position ):
            self.position= position
        else:
            raise TypeError( "position (%r) is an unrecognized type (%s)" % ( position, type(position) ) )
        if isinstance(qualifier, (list,tuple)):
            self.qualifier= qualifier
        else:
            self.qualifier= ( qualifier, )
        self.x12type= x12type
    def __repr__( self ):
        """Provide Documentation for epydoc."""
        typeName = "None" if self.x12type is None else self.x12type.__name__
        return "ElementAccess( %r, %r, %r, %s )" % ( self.segment, self.position, self.qualifier, typeName )

    def get_qualifier(self, instance, owner):
        """ Allow instance to influence the qualifier.
        
        This is a huge, regrettable hack. This will cause bugs that will take
        you hours to track down. I'm so sorry.

        I promise I'll come back some day and make all of this more reasonable,
        but I need to get this running asap.
        """
        if self.qualifier is None or self.qualifier == (None,):
            if hasattr(instance, 'qualifier'):
                return instance.qualifier
        else:
            return self.qualifier
        return (None,)

    def __get__( self, instance, owner ):
        qualifier = self.get_qualifier(instance, owner)
        if isinstance(instance, X12LoopBridge):
            segBridge= instance.segment( self.segment, qualifier[0], inList=qualifier[1:] )
        else:
            # ElementAccess works on a Segment which doesn't include qualifier
            # info, so make sure the current Segment has the qualifier 
            if qualifier[0] is None or \
                    instance.segment.getByPos(qualifier[0]) == qualifier[1]:
                segBridge = instance
            else:
                segBridge = None
        if segBridge is None:
            raw = None
        else:
            raw = self.position.get( segBridge.segment )
        if self.x12type is not None:
            return self.x12type.x12_to_python( raw )
        else:
            return raw
    def __set__( self, instance, value ):
        qualifier = self.get_qualifier(instance, owner)
        if x12type is not None:
            raw= self.x12type.python_to_x12( value )
        else:
            raw= value
        segBridge= instance.segment( self.segment, qualifier[0], inList=qualifier[1:] )
        if segBridge is None:
            raise MissingSegment( "Segment %s (%r) Not Found" % ( self.segment, qualifier ) )
        else:
            self.position.set( segBridge.segment, raw )

class ElementSequenceAccess( ElementAccess ):
    """Map a user-friendly attribute name to an sequence of Elements that occur in multiple
    copies of a Segment type.  This is used for the "NTE" Segments where
    a sequence of individual instances occur within a 2300 claim details loop.

    This is a Descriptor which implements the attributes name using
    __get__ and __set__ methods.
    """
    def __init__( self, segment, position, qualPos=None, inList=None, notInList=None, x12type=None ):
        """Build Element Access for a sequence of Elements.
        
        :param segment: Name of the segments that contains the elements which are the
        values for this attribute.
        :param position: integer position of the element within the segment.
        :param qualPos: integer position of the qualifier attribute in each segment instance.
        If omitted, all instances of the segment will become the sequence of values.
        If used, either :meth:`inList` or :meth:`notInList` must be provided.
        :param inList: sequence of values to include segments.
        :param notInList: sequence of values to exclude segments.
        :param x12type: An X12type conversion class name.
        :returns: The sequence of element values for this attribute name.
        """
        self.segment= segment
        self.position= position
        self.qualPos= qualPos
        self.inList= inList
        self.notInList= notInList
        self.x12type= x12type
    def __repr__( self ):
        """Provide Documentation for epydoc."""
        typeName = "None" if self.x12type is None else self.x12type.__name__
        return "ElementSequenceAccess( %r, %r, %r, %s )" % ( self.segment, self.position, self.qualPos, typeName )
    def __get__( self, instance, owner ):
        segs= instance.segList( self.segment, self.qualPos, self.inList, self.notInList )
        raw_list = [ s.segment.getByPos(self.position) for s in segs ]
        if self.x12type is not None:
            return [self.x12type.x12_to_python( raw ) for raw in raw_list]
        else:
            return raw_list
    def __set__( self, instance, valueList ):
        segs= instance.segList( self.segment, self.qualPos, self.inList, self.notInList )
        for seg,val in zip(segs, valueList):
            if seg is None or val is None:
                raise NotImplementedError( "Unimplemented: number of values and segments don't match" )
            if x12type is not None:
                raw= self.x12type.python_to_x12( val )
            else:
                raw= val
            seg.setByPos( self.position, raw )

class CompositePosition( object ):
    """Sets or gets the value of an Element in a Composite based on a
    single, fixed position.

    Segments are managed as SegmentToken instances.  Composites, however,
    are simple sequences of strings.
    """
    def __init__( self, *position ):
        self.positions= position
    def get( self, aComposite ):
        return [ aComposite[c] if c < len(aComposite) else None for c in self.positions ]
    def set( self, aComposite, *value ):
        for pos,val in zip(self.positions,value):
            assert pos is not None and val is not None, ValueError("values and composite positions diffent lengths")
            aComposite[pos]= val

class CompositeAccess( ElementAccess ):
    """Map a user-friendly attribute name to an Element of a Composite;
    the composites can be found
    anywhere among the occurances of a given Segment type.
    This is used for the various "HI" Segments where a single value is located in a
    qualified Composite.  This behaves a little bit like the OneOf position
    for an ordinary ElementAccess.

    Generally, a Composite has multiple Elements, the first of which is a qualifier.
    The rest of which are the target values.

    This is a Descriptor which implements the attributes name using
    __get__ and __set__ methods.  This descriptor is associated with a subclass of
    X12LoopBridge, the instance is a specific **Brige** and the owner is
    the subclass of X12LoopBridge.
    """
    def __init__( self, segment, compositeQualifier, compositePosition, x12type=None ):
        """Build Composite Access for an individual Composite.
        
        :param segment: Name of the segments that contains the composites which are the
        values for this attribute.
        :param compositeQualifier: A qualifier string or tuple of strings that identifies
        the composites to collect
        :param compositePosition: The numeric position of the element within the Composite.
        :param x12type: An X12type conversion class name.  This conversion will receive a
        large, complex sequence of values decomposed from a Composite.
        :returns: The sequence of element values for this attribute name.
        """
        self.segment= segment
        if isinstance( compositeQualifier, (list,tuple) ):
            self.qualifier= compositeQualifier
        else:
            self.qualifier= (compositeQualifier,)
        if isinstance( compositePosition, CompositePosition ):
            self.compPosition= compositePosition
        if isinstance( compositePosition, (list,tuple) ):
            self.compPosition= CompositePosition( *compositePosition )
        else:
            self.compPosition= CompositePosition( compositePosition )
        self.x12type= x12type
    def __repr__( self ):
        """Provide Documentation for epydoc."""
        typeName = "None" if self.x12type is None else self.x12type.__name__
        return "CompositeAccess( %r, %r, %r, %s )" % ( self.segment, self.position, self.qualifier, typeName )
    def __get__( self, instance, owner ):
        segList= instance.segList( self.segment )
        compList= []
        for seg in segList:
            compList.extend( seg.compositeList( *self.qualifier ) )
        data= []
        for composite in compList:
            raw= self.compPosition.get( composite )
            if self.x12type is not None:
                data.append( self.x12type.x12_to_python( raw ) )
            else:
                data.append( raw )
        return data[0]
    def __set__( self, instance, value ):
        """XXX - Implement this."""
        raise NotImplementedError( "Unimplemented: can't update Composite" )
        segList= instance.segList( self.segment )
        # need to locate composite in this segList and update it.

class CompositeSequenceAccess( CompositeAccess ):
    """Map a user-friendly attribute name to a sequence of Composites that occur
    somewhere in the occurances of a given Segment type
    This is used for the various "HI" Segments where a sequence of values is located in a
    qualified Composites.

    Each Composite Element has multiple values, the first of which is a qualifier
    for the Composite.  The rest of which is the target values.

    This is a Descriptor which implements the attributes name using
    __get__ and __set__ methods.
    """
    def __init__( self, segment, compositeQualifier=None, x12type=None ):
        """Build Composite Access for a sequence of Composites.
        
        :param segment: Name of the segments that contains the elements which are the
        values for this attribute.
        :param compositeQualifier: A qualifier string or tuple of strings that identifies
        the composites to collect
        :param x12type: An X12type conversion class name.  This conversion will receive a
        large, complex sequence of values decomposed from a Composite.
        :returns: The sequence of element values for this attribute name.
        """
        self.segment= segment
        if isinstance( compositeQualifier, (list,tuple) ):
            self.qualifier= compositeQualifier
        else:
            self.qualifier= ( compositeQualifier, )
        self.x12type= x12type
    def __repr__( self ):
        """Provide Documentation for epydoc."""
        typeName = "None" if self.x12type is None else self.x12type.__name__
        return "CompositeSequenceAccess( %r, %r, %r, %s )" % ( self.segment, self.position, self.qualifier, typeName )
    def __get__( self, instance, owner ):
        segList= instance.segList( self.segment )
        compList= []
        for seg in segList:
            compList.extend( seg.compositeList( *self.qualifier ) )
        data= []
        for composite in compList:
            raw= composite
            if self.x12type is not None:
                data.append( self.x12type.x12_to_python( raw ) )
            else:
                data.append( raw )
        return data
    def __set__( self, instance, value ):
        """XXX - implement this."""
        raise NotImplementedError( "Unimplemented: can't update Composite" )

class Conversion( object ):
    """Convert between X12 strings and Python objects.
    This is the abstract superclass for all conversions.
    """
    @staticmethod
    def x12_to_python( raw ):
        return NotImplemented
    @staticmethod
    def python_to_x12( value ):
        return NotImplemented


class TM(Conversion):
    """Convert between TM format time to proper datetime.time objects."""
    @staticmethod
    def x12_to_python(raw):
        if raw is None or raw == "":
            return raw
        if len(raw) < 4:
            return ""
        hh, mm = int(raw[0:2]), int(raw[2:4])
        ss = 0
        dd = 0  # hundredths of a second / centiseconds
        if len(raw) >= 6:
            ss = int(raw[4:6])
        if len(raw) >= 7:
            dd = int(raw[6:])
        dd = dd * 10  # milliseconds
        dd = dd * 100  # microseconds
        return datetime.time(hh, mm, ss, dd, None)

    @staticmethod
    def python_to_x12(value):
        if value is None:
            return ""
        return "{time}{centiseconds}".format(
                time=value.strftime("%H%M%S"),
                centiseconds=int(value.microsecond / 10000))


class D8( Conversion ):
    """Convert between D8 format dates to proper DateTime objects."""
    @staticmethod
    def x12_to_python( raw ):
        if raw is None:
            return raw
        yy,mm,dd = int(raw[0:4]), int(raw[4:6]), int(raw[6:8])
        return datetime.date( yy,mm,dd )
    @staticmethod
    def python_to_x12( value ):
        if value is None:
            return ""
        return value.strftime( "%4Y%2m%2d" )


class DR( Conversion ):
    """Convert between DR format dates to proper DateTime objects."""
    @staticmethod
    def x12_to_python( raw ):
        if raw is None or raw == "":
            return None
        d1, punct, d2 = raw.partition('-')
        if d1 is None or d2 is None or d1 == "" or d2 == "":
            return None
        yy1,mm1,dd1 = int(d1[0:4]), int(d1[4:6]), int(d1[6:8])
        yy2,mm2,dd2 = int(d2[0:4]), int(d2[4:6]), int(d2[6:8])
        return datetime.date( yy1,mm1,dd1 ), datetime.date( yy2,mm2,dd2 )
    @staticmethod
    def python_to_x12( value ):
        if value is None:
            return ""
        d1, d2 = value
        return "%s-%s" % ( d1.strftime( "%4Y%2m%2d" ), d2.strftime( "%4Y%2m%2d" ) )


class SegmentConversion( Conversion ):
    __name__ = "SegmentConversion"
    """Convert between an X12Segment and a proper Python object."""
    def __init__( self, someClass ):
        self.someClass= someClass

    def x12_to_python( self, raw ):
        return self.someClass( raw )

    def python_to_x12( self, value ):
        return value.segment


class XDecimal(Conversion):
    @staticmethod
    def x12_to_python(raw):
        if raw == "" or raw is None:
            return Decimal(0.0)
        if "(" in raw:
            raw = raw.replace("(", "-").replace(")", "")
        return Decimal(raw)

    @staticmethod
    def python_to_x12(value):
        if isinstance(value, str):
            return value
        return "{}".format(value)


class Money(XDecimal):
    @staticmethod
    def python_to_x12(value):
        if isinstance(value, str):
            return value
        return "{:.2f}".format(value)


def enum(options, raw_unknowns=False):
    """Translate abbreviations or codes into descriptions.

    :param options: A dictionary of code:description
    :type options: dict
    :param raw_unknowns: Whether or not to return the raw value if it's not
    in the options dict.
    :type raw_unknowns: boolean

    For example:

    class Header(X12LoopBridge):
        loopName = "HEADER"
        type = ElementAccess("BPR", 1, x12type=enum(
            {"C": "Payment accompanies remittance advice",
             "D": "Make payment only",
             "H": "Notification only",
             "I": "Remittance information only",
             "P": "Prenotification of future transfers",
             "U": "Split payment and remittance",
             "X": "Handling party's option to split payment and remittance."},
             raw_unknowns=True))
    header.type -> ("C", "Payment accompanies remittance advice")

    # Or an unknown type, say "Z":
    header.type -> "Z"
    """
    class Enum(Conversion):
        @staticmethod
        def x12_to_python(raw):
            if raw == "" or raw is None:
                return None
            elif raw in options:
                return (raw, options[raw])
            elif raw_unknowns:
                return raw
            return None

        @staticmethod
        def python_to_x12(value):
            if value is None or value == "":
                return ""
            elif value in options:
                return value
            else:
                for (k,v) in options.iteritems():
                    if value == v:
                        return k
            if raw_unknowns:
                return value
            return ""
    return Enum


def boolean(match):
    """ True if the field matches match. Match must be a string. """
    class Boolean(Conversion):
        @staticmethod
        def x12_to_python(raw):
            return raw == match

        @staticmethod
        def python_to_x12(value):
            return match
    return Boolean



#!/usr/bin/env python
"""Convert pyx12 configuration files from XML to Python.

The Python X12 module contains a number of XML files which
define X12 message structures.  They need to be converted into the syntax for the
:mod:`X12.parse` parser.

Note that the pyx12 module defines Loops, Segments, Composites and individual Elements.
It also defines selected data element types and sizes.  Since it provides
complete code values, it provides Segment parsing clues and compliance
checking values.

Definitions
===========

There are two principle classes in this module.

..  autoclass:: XMLParser

    This reads XML files and creates
    :mod:`xml.dom` structures for the codes, data elements and message
    structures.

..  autoclass:: ParserBuilder

    This emits a Python :mod:`X12.parse` constructor
    that matches the given :file:`.CF` file.

There are two convenience functions.

..  autofunction:: convertFilePath

..  autofunction:: convertFile

..  autofunction:: convertAll


Warnings
========

The following default warnings configuration is used.

    -   :class:`XMLWarning` are elevated to Errors to cause this module to stop.
        In the case of suspicious XML, it is possible to downgrade these
        errors and keep processing.

    -   :class:`EmptyLoopWarning` is always reported, but is merely a warning.

    -   :class:`UnknownElementWarning` is always reported, but is merely a warning.

    -   :class:`Extension` is reported once so that it's clear that the XML contains
        features which were not parsed.

XML Source Structure
====================

The pyx12 package provides a :mod:`pyx12.maps.xsd` that provides a bit
of guidance on parsing the XML documents.
The :file:`README` provides some additional hints.

The following elements are
used in the message xml definitions.

    - usage
        Implementation Guide Usage field. S=Situational, N=Not Used, R=Required

    - pos
        Segment position number -- obvious from the XML document, but provided anyway

    - id
        Implementation Guide id

    - name
        Implementation Guide name

    - data_ele
        Data Element id

    - valid_codes
        List of valid codes for the element or the name.
        Could be a separate list document, via attribute external="".
        See codes.xml for the complete set of codes.

    - refdes
        reference designator for a Composite item

    - syntax
        STD syntax dependency string for a segment, optional

The :mod:`pyx12.maps.xsd` provides a schema for the various message definitions.

Data Element Definitions
========================

The following are used in the data element definitions.

    - data_ele
        Data Element ID

    - name
        IG name

    - data_type
        STD type

    - min_len/max_len
        STD

Analysis reveals the following DTD for the message definitions::

    <!ELEMENT data_elements ( data_ele )*>

    <!ELEMENT data_ele #PCDATA>
    <!ATTLIST data_ele
        ele_num   CDATA #REQUIRED
        data_type (ID|R|AN|DT|TM|N0) #REQUIRED
        min_len   CDATA #REQUIRED
        max_len   CDATA #REQUIRED
        name      CDATA #REQUIRED>

"""
from __future__ import print_function
import xml.dom.minidom as DOM
import argparse
from datetime import datetime
import os.path, logging, sys, fnmatch
import warnings
from tigershark.X12.parse import Composite
from tigershark.X12.parse import Element
from tigershark.X12.parse import Loop
from tigershark.X12.parse import Message
from tigershark.X12.parse import Properties
from tigershark.X12.parse import Segment
from tigershark.X12.map.source import FlatPythonVisitor
from tigershark.X12.map.source import PythonVisitor

class XMLWarning( UserWarning ):
    """A superclass for a number of warnings regarding the XML definition."""
    pass

class EmptyLoopWarning( UserWarning ):
    """A Loop definition had no subloops or Segments."""
    pass

class UnknownElementWarning( UserWarning ):
    """An Element had a dataele= reference to an unkown name."""
    pass

class Extension( UserWarning ):
    """This feature of the XML is an extension to the current implementation."""
    pass

class ParserBuilder( object ):
    """Build a :mod:`X12.parse` Parser from an XML message specification.

    This makes use of an :class:`XMLParser` instance which has been populated
    with three separate XML documents.

        1. The data element definitions were parsed with :meth:`XMLParser.data`.

        2. The external code definitions were parsed with :meth:`XMLParser.codes`.

        3. The message structure was parsed via :meth:`XMLParser.read`.

    Once this is complete, then the :meth:`build` method can be given the
    :class:`XMLParser` to emit the desired :class:`X12.parse.Message` object.

    This creates the :class:`X12.parse.Message` object, inserts the various
    Loop and Segment components.  The resulting object can be used to
    parse X12 messages.

    The resulting object can also be used to emit Python source,
    as well as SQL DDL.

    The :samp:`<transaction>` contains :samp:`<loop>`'s.
    The :samp:`<loop>`'s contain :samp:`<segment>`'s and :samp:`<loop>`'s.
    The :samp:`<segment>`'s contain :samp:`<element>`'s and :samp:`<composite>`'s.

    :ivar dataDictionary: a dictionary of :samp:`<data_ele>` definitions.
    """
    def __init__( self ):
        """Creates a new, uninitialized ParserBuilder."""
        self.log= logging.getLogger( "tools.convertPyX12.ParserBuilder" )
        self.log2= logging.getLogger( "tools.convertPyX12.ParserBuilder.buildElement" )
        self.dataDictionary= {}
    def buildDataElementDef( self, aNode ):
        """Load the data dictionary with the definition from a :samp:`<data_ele>` node.
        The key is the data element name.
        The value in the dictionary is a tuple of ( data_type, min_len and max_len ).

        :param aNode: a :mod:`xml.dom` Element with a name of :samp:`data_ele`
        """
        assert aNode.nodeType == DOM.Node.ELEMENT_NODE and aNode.nodeName == "data_ele"
        # ele_num="100" data_type="ID" min_len="3" max_len="3" name
        ele_num= aNode.getAttribute('ele_num')
        data_type= aNode.getAttribute('data_type')
        min_len= aNode.getAttribute('min_len')
        max_len= aNode.getAttribute('max_len')
        name= aNode.getAttribute('name')
        self.dataDictionary[ele_num]= ( name, (data_type, min_len, max_len) )
    def dataElements( self, xmlDoc ):
        """Load the data dictionary with the definitions from the :samp:`<data_elements>` node.

        :param xmlDoc: a :class:`XMLParser` which has parsed the data document
        """
        doc= xmlDoc.dataeleDoc.documentElement
        assert doc.nodeType == DOM.Node.ELEMENT_NODE and doc.nodeName == "data_elements"
        for c in doc.childNodes:
            if c.nodeType != DOM.Node.ELEMENT_NODE: continue
            if c.nodeName == "data_ele":
                self.buildDataElementDef( c )
            else:
                warnings.warn( XMLWarning("Unexpected %r" % (c,) ) )
                self.log.warning( "*** Unexpected %r", c )
    def codes( self, xmlDoc ):
        """Load external code file.

        ..  todo:: XXX - Finish load external code file.

        :param xmlDoc: a :class:`XMLParser` which has parsed the codes document
        """
        pass
    def getValidCodes( self, aNode ):
        """Parse a :samp:`<valid_codes>` node, building a list of valid codes
        for an Element.  This will examine only :samp:`<code>` elements found
        under the :samp:`<valid_codes>` element.
        
        :param aNode: a :mod:`xml.dom` Element with a name of :samp:`valid_codes`
        :returns: list of code values
        """
        assert aNode.nodeType == DOM.Node.ELEMENT_NODE and aNode.nodeName == "valid_codes"
        codes= []
        for c in aNode.childNodes:
            if c.nodeType != DOM.Node.ELEMENT_NODE: continue
            if c.nodeName == "code":
                for n in c.childNodes:
                    codes.append( n.nodeValue )
            else:
                warnings.warn( XMLWarning("Unexpected %r" % (c,) ) )
                self.log.warning( "*** Unexpected %r", c )
        return codes
    def buildComposite( self, compositeNode, context, nesting=0 ):
        """Build a Composite from a list of sub-Elements.

        Note that the ISA segment provides the Component Element Separator in ISA16.

        Composites are tricky.  We have two choices.

            1.  Treat them as just a sequence of Elements -- which they
                are for RDBMS purposes.  However, this makes compliance
                checking hard, since we have to look inside the composite.
                We have to turn off "match" for each Element of the Composite,
                otherwise, we won't be able to look into the segment data
                properly.

            2.  Treat them as a subclass of Elements -- which is a lot of
                fooling around to support a few compliance checks.  We have to
                provide a composite match function and we have to make use
                of punctuation in the ISA segment for sub-parsing each Composite when we
                finally figure out what the Segment and Elements.

        :param compositeNode: a :mod:`xml.dom` Element with a name of :samp:`composite`
        :param context: The Segment which contains this Composite
        :param nesting: The current nesting level used to indent the log messages.
        """
        assert compositeNode.nodeType == DOM.Node.ELEMENT_NODE and compositeNode.nodeName == "composite"
        name= self.getChildTextValue( compositeNode, "name" )
        usage= self.getChildTextValue( compositeNode, "usage" )
        seq= self.getChildTextValue( compositeNode, "seq" )
        data_ele= self.getChildTextValue( compositeNode, "data_ele" )
        refdes= self.getChildTextValue( compositeNode, "refdes" )
        self.log.debug( "%*sComposite name %r usage %r seq %r data_ele %r refdes %r",
                       nesting*2, '', name, usage, seq, data_ele, refdes )
        theComposite= Composite(
            data_ele,
            Properties( desc=name, req_sit=usage, seq=seq, refdes=refdes ) )
        for c in compositeNode.childNodes:
            # Want to preserve the original XML order of <element>
            if c.nodeType != DOM.Node.ELEMENT_NODE: continue
            if c.nodeName == "element":
                self.buildElement( c, theComposite, nesting+1 )
            elif c.nodeName in ( "name", "usage", "seq", "data_ele", 'refdes',):
                pass # already consumed
            else:
                warnings.warn( XMLWarning("Unexpected %r" % (c,) ) )
                self.log.warning( "*** Unexpected %r", c )
        context.append( theComposite )
    def buildElement( self, elementNode, context, nesting=0 ):
        """Element is the fundamental piece of data in a Segment or Composite.
        Elements can have code definitions or can reference external code
        definitions.
        
        ..  todo:: XXX - Use external code definitions.
        
        :param elementNode: a :mod:`xml.dom` Element with a name of :samp:`element`
        :param context: The Segment or Composite which contains this Element
        :param nesting: The current nesting level used to indent the log messages.
        """
        assert elementNode.nodeType == DOM.Node.ELEMENT_NODE and elementNode.nodeName == "element"
        eltXid= elementNode.getAttribute('xid')
        name= self.getChildTextValue( elementNode, "name" )
        usage= self.getChildTextValue( elementNode, "usage" )
        seq= self.getChildTextValue( elementNode, "seq" )
        data_ele= self.getChildTextValue( elementNode, "data_ele" )
        self.log.debug( "%*sElement id %r name %r usage %r seq %r data_ele %r",
                       nesting*2, '', eltXid, name, usage, seq, data_ele )
        if not self.dataDictionary.has_key( data_ele ):
            warnings.warn( UnknownElementWarning("No Data Element %r %r %r" % ( eltXid, data_ele, name ,) ) )
            data_type_tuple= (None,None,None)
        else:
            name, data_type_tuple = self.dataDictionary[data_ele]

        codes= []
        for c in elementNode.childNodes:
            if c.nodeType != DOM.Node.ELEMENT_NODE: continue
            if c.nodeName == "valid_codes":
                if c.getAttribute("external") != "":
                    pass # reference to codes.xml codes
                    # XXX - Handle external code list
                    warnings.warn( Extension("External Codes Not Implemented") )
                else:
                    codes= self.getValidCodes( c )
            elif c.nodeName in ( "name", "usage", "seq", "data_ele", ):
                pass # already consumed
            else:
                warnings.warn( XMLWarning("Unexpected %r" % (c,) ) )
                self.log.warning( "*** Unexpected %r", c )

        theElement= Element(
            eltXid,
            Properties(desc=name, req_sit=usage, seq=seq, data_ele=data_ele,
                       data_type=data_type_tuple, codes=codes, )
            )

        if self.dataDictionary.has_key( data_ele ) and len(codes) != 0:
            #self.log2.debug( "Segment Qual Parameter? %r %r", self.dataDictionary[data_ele], codes )
            pass

        context.append( theElement )
    def buildSegment( self, segmentNode, context, nesting=0 ):
        """Segment contains Elements and Composites.
        
        :param segmentNode: a :mod:`xml.dom` Element with a name of :samp:`segment`
        :param context: The Loop which contains this Segment.
        :param nesting: The current nesting level used to indent the log messages.
        """
        assert segmentNode.nodeType == DOM.Node.ELEMENT_NODE and segmentNode.nodeName == "segment"
        segXid= segmentNode.getAttribute('xid')
        name= self.getChildTextValue( segmentNode, "name" )
        usage= self.getChildTextValue( segmentNode, "usage" )
        pos= self.getChildTextValue( segmentNode, "pos" )
        max_use= self.getChildTextValue( segmentNode, "max_use" )
        syntax= self.getChildTextValue( segmentNode, "syntax" )
        self.log.debug( "%*sSegment xid %r: name %r usage %r pos %r max_use %r syntax %r",
                       nesting*2, '', segXid, name, usage, pos, max_use, syntax )
        theSegment= Segment(
            segXid,
            Properties(desc=name,req_sit=usage,pos=pos,repeat=max_use,syntax=syntax),
            )

        for c in segmentNode.childNodes:
            # Want to preserve the original XML order of <element> and <composite>
            if c.nodeType != DOM.Node.ELEMENT_NODE: continue
            if c.nodeName == "element":
                self.buildElement( c, theSegment, nesting+1 )
            elif c.nodeName == "composite":
                self.buildComposite( c, theSegment, nesting+1 )
            elif c.nodeName in ( "name", "usage", "pos", "max_use", "syntax", ):
                pass # already consumed
            else:
                warnings.warn( XMLWarning("Unexpected %r" % (c,) ) )
                self.log.warning( "*** Unexpected %r", c )
        context.append( theSegment )
    def buildLoop( self, loopNode, context, nesting=0 ):
        """Loop contains Segments and Loops.
        Empty Loops create a warning, and are dropped.
        
        :param loopNode: a :mod:`xml.dom` Element with a name of :samp:`loop`
        :param context: The Loop which contains this Loop or Segment.
        :param nesting: The current nesting level used to indent the log messages.
        """
        assert loopNode.nodeType == DOM.Node.ELEMENT_NODE and loopNode.nodeName == "loop"
        loopXid= loopNode.getAttribute('xid')
        loopType= loopNode.getAttribute('type')
        name= self.getChildTextValue( loopNode, "name" )
        usage= self.getChildTextValue( loopNode, "usage" )
        pos= self.getChildTextValue( loopNode, "pos" )
        repeat= self.getChildTextValue( loopNode, "repeat" )
        self.log.debug( "%*sLoop xid %r type %r: name %r usage %r pos %r repear %r",
                       nesting*2, '', loopXid, loopType, name, usage, pos, repeat )
        theLoop= Loop(
            loopXid,
            Properties(desc=name,req_sit=usage,pos=pos,repeat=repeat,looptype=loopType),
            )

        for c in loopNode.childNodes:
            # Want to preserve the original XML order of <loop> and <segment>
            if c.nodeType != DOM.Node.ELEMENT_NODE: continue
            if c.nodeName == "loop":
                self.buildLoop( c, theLoop, nesting+1 )
            elif c.nodeName == "segment":
                self.buildSegment( c, theLoop, nesting+1 )
            elif c.nodeName in ( "name", "usage", "pos", "repeat", ):
                pass # already consumed
            else:
                warnings.warn( XMLWarning("Unexpected %r" % (c,) ) )
                self.log.warning( "*** Unexpected %r", c )
        if len(theLoop.structure) == 0:
            warnings.warn( EmptyLoopWarning("Empty Loop %r" % (theLoop,) ) )
            # optimize this out of existence
        else:
            context.append( theLoop )
    def build( self, xmlDoc, name=None ):
        """Build the overall :class:`X12.parse.Message` parser.
        
        :param xmlDoc: a :class:`XMLParser` which has parsed the data, codes and message
            structure documents.
        :param name: Optional name of the Message to build; this will use
            the xid attribute provided in the XML if no overriding name is provided here.
        :returns: :class:`X12.parse.Message` parser.
        """
        self.dataElements( xmlDoc )
        self.codes( xmlDoc )
        doc= xmlDoc.doc.documentElement
        assert doc.nodeType == DOM.Node.ELEMENT_NODE and doc.nodeName == "transaction"
        xid= doc.getAttribute('xid')
        desc= self.getChildTextValue( doc, "name" )
        if name is None: name= xid
        self.log.info( "Message %s: xid=%s desc=%s", name, xid, desc )
        self.top= Message( name, Properties(desc=desc) )
        for c in doc.childNodes:
            # Want to preserve the original XML order of <loop> and <segment>
            if c.nodeType != DOM.Node.ELEMENT_NODE: continue
            if c.nodeName == "loop":
                self.buildLoop( c, self.top )
            elif c.nodeName in ( "name", ):
                pass # Already consumed this
            else:
                warnings.warn( XMLWarning("Unexpected %r" % (c,) ) )
                self.log.warning( "*** Unexpected %r", c )
        return self.top
    def getChildTextValue( self, aNode, name ):
        """Examines all children with the given name
        and extracts the text nodes for those children.
        It accumulates the nodeValues for those text nodes.
        
        :param aNode: an :mod:`xml.dom` Element.
        :param name: a node name underneath the given node.
        """
        childElements = [ n for n in aNode.childNodes if n.nodeType == DOM.Node.ELEMENT_NODE and n.nodeName == name ]
        textNodes = [ n for c in childElements for n in c.childNodes if n.nodeType in ( DOM.Node.TEXT_NODE, DOM.Node.CDATA_SECTION_NODE ) ]
        text = [ n.nodeValue for n in textNodes ]
        return " ".join( text )

class XMLParser( object ):
    """Parse a related set of XML docs to get codes, data element definitions
    and the overall Message structure.

    The ParserBuilder will need some combination of these related XML documents
    to define the :mod:`X12.parse` parser.
    """
    def __init__( self ):
        """Creates an empty XMLParser instance."""
        self.dataeleDoc= None
        self.codesDoc= None
        self.doc= None
    def data( self, name_or_file ):
        """Parses the data element definition XML file.
        
        :param fileName: name of the data element file, usually :file:`dataele.xml`
        """
        self.dataeleDoc= DOM.parse( name_or_file )
    def codes( self, name_or_file ):
        """Parses the external codes definition XML file.
        
        :param fileName: name of the external codes file, usually :file:`codes.xml`
        """
        self.codesDoc= DOM.parse( name_or_file )
    def read( self, name_or_file ):
        """Parses the message structure XML file.
        
        :param fileName: the name of a message structure file.
        """
        self.doc= DOM.parse( name_or_file )
    def dump( self, aDoc ):
        """Prints a dump of an XML document to stdout.
        
        :param aDoc: an :mod:`xml.dom` Node (usually the top-level Document).
        """
        nodeNames= set()
        def walk( node, depth=0 ):
            if node.nodeType in ( DOM.Node.TEXT_NODE, DOM.Node.CDATA_SECTION_NODE ):
                if len(node.nodeValue.strip()) == 0:
                    # Disposable whitespace node
                    return
            if node.nodeType == DOM.Node.COMMENT_NODE:
                return
            print( depth*' ', node.nodeName, repr(node.nodeValue) )
            if node.attributes is not None and len(node.attributes.keys()) > 0:
                print( depth*'  ', '(', ", ".join([ "%s=%r" % (k,v) for k,v in node.attributes.items() ]), ')' )
            nodeNames.add( node.nodeName )
            for c in node.childNodes:
                walk( c, depth+1 )
        walk( aDoc )

def convertFilePath( baseDir, aFile, dataeleFile="dataele.xml",
        codesFile="codes.xml" ):
    """Converts a single message file from XML to :mod:`X12.parse`."""
    return convertFile(open(os.path.join(baseDir, aFile)),
            open(os.path.join(baseDir, dataeleFile)),
            open(os.path.join(baseDir, codesFile)))


def convertFile(xml_file, data_ele_file, codes_file):
    log = logging.getLogger("tools.convertPyX12.convertFile")
    bp = ParserBuilder()
    try:
        xml = XMLParser()
        xml.data(data_ele_file)
        xml.codes(codes_file)
        xml.read(xml_file)
        return bp.build(xml)
    except Exception as e:
        log.error("Failed to convert %s: %s" % (xml_file.name, e))
        raise


def convertAll( baseDir ):
    """Convert all :file:`nnn*.4010.X*.xml` files in the given directory.
    
    :param baseDir: Directory with message definition, dataele and codes files.
    """
    for path,dirs,names in os.walk(baseDir):
        for fn in names:
            if fnmatch.fnmatch( fn, "[0-9][0-9][0-9]*.4010.X*.xml"):
                convertFilePath( path, fn )

def writeFile(aFile, name, x12, structure="flat"):
    """Write the x12 python module to a file.
    
    :param aFile: Filename of destination file.
    :type aFile: String
    :param name: The name of the generated class.
    :type name: String
    :param x12: The `X12.parse.Message` object to write.
    :type x12: `X12.parse.Message`
    """
    if structure == "nested":
        pyMap= PythonVisitor( name )
    else:
        pyMap= FlatPythonVisitor( name )
    x12.visit( pyMap )
    pySrc= pyMap.getSource()
    with open(aFile, 'w') as f:
        f.write("#\n# Generated by TigerShark.tools.convertPyX12 on %s\n#\n" % 
                datetime.now())
        f.write( pySrc )
        f.write( '\n\n' )


# Raise an XML warning to an error level.  This will spot extra tags.
warnings.simplefilter( "error", XMLWarning, 0 )
# Report empty loop warnings.  This will spot Loops with no Segments.
warnings.simplefilter( "always", EmptyLoopWarning, 0 )
# Report unknown data elements.
warnings.simplefilter( "always", UnknownElementWarning, 0 )
# Report extensions that aren't implemented yet.
warnings.simplefilter( "once", Extension, 0 )

if __name__ == "__main__":
    logging.basicConfig( stream=sys.stdout, level=logging.DEBUG )
    parser = argparse.ArgumentParser(
        description="Convert a PyX12 XML file to a python module.")
    parser.add_argument('x12_file', help="The x12 xml file to convert")
    parser.add_argument('py_file', help="The destination file.")
    parser.add_argument('-b', '--base_dir', dest="base_dir", default='.',
            help="Base directory containing the X12 xml, dataele.xml, and "\
                    "codes.xml files.")
    parser.add_argument('-s', '--structure', choices=['flat', 'nested'],
            default="flat", help="The structure of the resulting python "\
                    "class. Nested is easier to read, but may not compile "\
                    "due to too many object instantiations in a single "\
                    "call.")
    parser.add_argument('-n', '--name', help="The name of the generated "\
            "class. Defaults to the py_file argument, minus the filetype.")
    args = parser.parse_args()
    x12 = convertFilePath( args.base_dir, args.x12_file)
    if args.name is not None:
        name = args.name
    else:
        name = args.py_file.rsplit('/', 1)[1]
        name = name.split('.', 1)[0]
    writeFile(args.base_dir, args.py_file, name, x12, args.structure)



from tigershark.facade import D8
from tigershark.facade import DR
from tigershark.facade import ElementAccess
from tigershark.facade import TM
from tigershark.facade import X12LoopBridge
from tigershark.facade import X12SegmentBridge
from tigershark.facade import SegmentAccess
from tigershark.facade import SegmentConversion
from tigershark.facade import boolean
from tigershark.facade import enum
from tigershark.facade.enums.common import date_or_time_qualifier
from tigershark.facade.enums.common import id_code_qualifier
from tigershark.facade.enums.common import reference_id_qualifier


class Header(X12LoopBridge):
    loopName = "HEADER"

    class _HierarchicalTransaction(X12SegmentBridge):
        """Use this to start the transaction set.

        This also indicates the sequence of hierarchical levels of information
        that will follow, though that information is not used at this time."""
        structure = ElementAccess("BHT", 1, x12type=enum(
            {"0022": "Information Source, Information Receiver, Subscriber, "
                "or Dependent."}, raw_unknowns=True))
        purpose = ElementAccess("BHT", 2, x12type=enum({
            "01": "Cancellation",
            "11": "Response",
            "13": "Request",
            "36": "Authority to Dedduct (Reply)"}, raw_unknowns=True))
        transaction_id = ElementAccess("BHT", 3)
        creation_date = ElementAccess("BHT", 4, x12type=D8)
        creation_time = ElementAccess("BHT", 5, x12type=TM)
        type = ElementAccess("BHT", 6, x12type=enum({
            "RT": "Spend Down",
            "RU": "Medical Services Reservation"}, raw_unknowns=True))

    hierarchical_transaction_info = SegmentAccess("BHT",
            x12type=SegmentConversion(_HierarchicalTransaction))


class Hierarchy(X12SegmentBridge):
    id = ElementAccess("HL", 1)
    parent_id = ElementAccess("HL", 2)
    level = ElementAccess("HL", 3, x12type=enum(
        {"20": "Information Source",
         "21": "Information Receiver",
         "22": "Subscriber",
         "23": "Dependent",
            }))
    has_children = ElementAccess("HL", 4, x12type=boolean("1"))


class TraceNumber(X12SegmentBridge):
    """ Uniquely identify this transaction set.

    Also aid in reassociating payments and remittances that have been
    separated.
    """
    trace_type = ElementAccess("TRN", 1, x12type=enum({
            "1": "Current Transaction Trace Numbers",
            "2": "Referenced Transaction Trace Numbers (Value from 270)"}))
    trace_number = ElementAccess("TRN", 2)
    entity_id = ElementAccess("TRN", 3)
    entity_additional_id = ElementAccess("TRN", 4)


class NamedEntity(X12SegmentBridge):
    entity_identifier = ElementAccess("NM1", 1, x12type=enum({
            "03": "Dependent",
            "1P": "Provider",
            "2B": "Third-Party Administrator",
            "36": "Employer",
            "80": "Hospital",
            "FA": "Facility",
            "GP": "Gateway Provider",
            "IL": "Insured",
            "P5": "Plan Sponsor",
            "PR": "Payer",
            "QC": "Patient"}))
    entity_type = ElementAccess("NM1", 2, x12type=enum({
            "1": "Person",
            "2": "Non-Person Entity"}))

    last_name = ElementAccess("NM1", 3)
    org_name = ElementAccess("NM1", 3)
    first_name = ElementAccess("NM1", 4)
    middle_initial = ElementAccess("NM1", 5)
    suffix = ElementAccess("NM1", 7)

    id_code = ElementAccess("NM1", 9)
    id_code_qual = ElementAccess("NM1", 8, x12type=enum(id_code_qualifier))

    @property
    def is_person(self):
        return self.entity_type[0] == "1"

    @property
    def is_organization(self):
        return self.entity_type[0] == "2"


class ReferenceID(X12SegmentBridge):
    reference_id_qualifier = ElementAccess("REF", 1,
            x12type=enum(reference_id_qualifier))
    reference_id = ElementAccess("REF", 2)
    description = ElementAccess("REF", 3)


class ContactInformation(X12SegmentBridge):
    contact_code = ElementAccess("PER", 1, x12type=enum({
        "IC": "Information Contact"}, raw_unknowns=True))
    contact_name = ElementAccess("PER", 2)
    contact_edi = ElementAccess("PER", oneOf=("ED", (3, 4), (5, 6), (7, 8)))
    contact_email = ElementAccess("PER", oneOf=("EM", (3, 4), (5, 6), (7, 8)))
    contact_fax = ElementAccess("PER", oneOf=("FX", (3, 4), (5, 6), (7, 8)))
    contact_home_phone = ElementAccess("PER",
            oneOf=("HP", (3, 4), (5, 6), (7, 8)))
    contact_work_phone = ElementAccess("PER",
            oneOf=("WP", (3, 4), (5, 6), (7, 8)))
    contact_phone = ElementAccess("PER", oneOf=("TE", (3, 4), (5, 6), (7, 8)))
    contact_phone_ext = ElementAccess("PER",
            oneOf=("EX", (3, 4), (5, 6), (7, 8)))


class Address(X12SegmentBridge):
    addr1 = ElementAccess("N3", 1)
    addr2 = ElementAccess("N3", 2)


class Location(X12SegmentBridge):
    city = ElementAccess("N4", 1)
    state = ElementAccess("N4", 2)
    zip = ElementAccess("N4", 3)
    country_code = ElementAccess("N4", 4)
    location_type = ElementAccess("N4", 5, x12type=enum({
        "CY": "County/Parish",
        "FI": "Federal Information Processing Standards (FIPS) 55 (Named "
                "Populated Places)"}, raw_unknowns=True))
    location_id = ElementAccess("N4", 6)


class DemographicInformation(X12SegmentBridge):
    birth_date = ElementAccess("DMG", 2, x12type=D8, qualifier=(1, "D8"))
    gender = ElementAccess("DMB", 3, x12type=enum({
        "F": "Female",
        "M": "Male",
        "U": "Unknown"}))


class Relationship(X12SegmentBridge):
    is_insured = ElementAccess("INS", 1, x12type=boolean('Y'))
    relationship = ElementAccess("INS", 2, x12type=enum({
        "01": "Spouse",
        "18": "Self",
        "19": "Child",
        "21": "Unknown",
        "34": "Other Adult"}, raw_unknowns=True))
    maintenance_type = ElementAccess("INS", 3, x12type=enum({
        "001": "Change"}, raw_unknowns=True))
    maintenance_reason = ElementAccess("INS", 4, x12type=enum({
        "25": "Change in Identifying Data Elements"}, raw_unknowns=True))
    student_status = ElementAccess("INS", 9, x12type=enum({
        "F": "Full-time",
        "N": "Not a Student",
        "P": "Part-time"}))
    handicapped = ElementAccess("INS", 10, x12type=boolean("Y"))
    birth_sequence_number = ElementAccess("INS", 17)


class DateOrTimePeriod(X12SegmentBridge):
    type = ElementAccess("DTP", 1, x12type=enum(date_or_time_qualifier))
    time = ElementAccess("DTP", 3, x12type=D8, qualifier=(2, "D8"))
    time_range = ElementAccess("DTP", 3, x12type=DR, qualifier=(2, "RD8"))


class Diagnosis(X12SegmentBridge):
    principal_diagnosis_icd9_code = ElementAccess("III", 2,
            qualifier=(1, "BK"))
    diagnosis_icd9_code = ElementAccess("III", 2, qualifier=(1, "BF"))
    other_code = ElementAccess("III", 2, qualifier=(1, "ZZ"), x12type=enum({
        "11": "Office",
        "12": "Home",
        "21": "Inpatient Hospital",
        "22": "Outpatient Hospital",
        "23": "Emergency Room - Hospital",
        "24": "Ambulatory Surgical Center",
        "25": "Birthing Center",
        "26": "Military Treatment Facility",
        "31": "Skilled Nursing Facility",
        "32": "Nursing Facility",
        "33": "Custodial Care Facility",
        "34": "Hospice",
        "41": "Ambulance - Land",
        "42": "Ambulance - Air or Water",
        "50": "Federally Qualified Health Center",
        "51": "Inpatient Psychiatric Facility",
        "52": "Psychiatric Facility Partial Hospitalization",
        "53": "Community Mental Health Center",
        "54": "Intermediate Care Facility/Mentally Retarded",
        "55": "Residential Substance Abuse Treatment Facility",
        "56": "Psychiatric Residential Treatment Center",
        "60": "Mass Immunization Center",
        "61": "Comprehensive Inpatient Rehabilitation Facility",
        "62": "Comprehensive Outpatient Rehabilitation Facility",
        "65": "End-Stage Renal Disease Treatment Facility",
        "71": "State or Local Public Health Clinic",
        "72": "Rural Health Clinic",
        "81": "Independent Laboratory 99 Other Unlisted Facility"},
        raw_unknowns=True))


class ProviderInformation(X12SegmentBridge):
    provider_code = ElementAccess("PRV", 1, x12type=enum({
        "AD": "Admitting",
        "AT": "Attending",
        "BI": "Billing",
        "CO": "Consulting",
        "CV": "Covering",
        "H": "Hospital",
        "HH": "Home Health Care",
        "LA": "Laboratory",
        "OT": "Other Physician",
        "P1": "Pharmacist",
        "P2": "Pharmacy",
        "PC": "Primary Care Physician",
        "PE": "Performing",
        "R": "Rural Health Clinic",
        "RF": "Referring",
        "SB": "Submitting",
        "SK": "Skilled Nursing Facility",
        "SU": "Supervising"}))
    reference_id_qualifier = ElementAccess("PRV", 2, x12type=enum({
        "9K": "Servicer",
        "D3": "National Association of Boards of Pharmacy Number",
        "EI": "Employer's Identification Number",
        "HPI": "Healthcare Financing Administration National Provider ID",
        "SY": "Social Security Number",
        "TJ": "Federal Taxpayer's Identification Number",
        "ZZ": "Mutually Defined"}, raw_unknowns=True))
    reference_id = ElementAccess("PRV", 3)



#!/usr/bin/env python
"""The X12.map.SQL module emits SQL table definitions from an :mod:`X12.parse` structure.

These table definitions are minimized by flattening Segments into their
parent Loop structures.

See :ref:`traversal` for notes on the **Visitor** design pattern.

..  autoclass:: SQLDetailsVisitor
    :members:

..  autoclass:: SQLTableVisitor
    :members:
"""
from tigershark.X12.parse import StructureVisitor
from tigershark.X12.parse import StopDescent

class SQLDetailsVisitor( StructureVisitor ):
    """Details means, effectively, column definitions, which are X12.parse.Elements.
    However, a Segments are murky.
    A Segment with a repeat of "1" can be expanded inline.
    A Segment with a repeat of ">1" or some specific number, has to be a separate
    table (or we have to generate a design not in 1NF).

    Similarly, a Loop which repeats exactly once can be expanded in-line.
    Loops which repeat, on the other hand, require distinct tables.  This is
    not as interesting, so we trust the other Visitor to allocate all Loops
    to separate tables.

    This visitor merely emits the in-line Segments, in-line Loops and Elements.
    Composites are little more than a comment, effectively an in-line sub-Segment.
    """
    typeMap= {
        'ID': 'VARCHAR2',
        'R':  'DECIMAL',
        'AN': 'VARCHAR2',
        'DT': 'CHAR', # Date, really
        'TM': 'CHAR', # Time, really
        'N0': 'NUMERIC',
        'N':  'NUMERIC',
        'B':  'NUMERIC', # Binary?
    }
    def __init__( self, *arg, **kw ):
        super( SQLDetailsVisitor, self, ).__init__( *arg, **kw )
        self.result= []
    def postElement( self, anElement, indent ):
        if anElement.req_sit == "S":
            nullable= "         "
        else:
            nullable= "NOT  NULL"
        # Some Segments are repeated.
        if anElement.parent.total_occurs != 1:
            qual= anElement.parent.occurance
        else:
            qual= ""
        typeStr= "%s(%s)" % ( self.typeMap.get(anElement.type_name,None), anElement.max_len )
        self.result.append( "    %s%s %s %s, -- %s" % (
            anElement.sqlName, qual, typeStr, nullable, anElement.desc ) )
    def preComposite( self, aComposite, indent ):
        self.result.append( "    -- Composite %s %s" % ( aComposite.name, aComposite.desc, ))
    def preSegment( self, aSegment, indent ):
        """Repeats?  If so, we need a separate table."""
        if aSegment.repeat == "1":
            self.result.append( "    -- In-line Segment %s %s" % (aSegment.name, aSegment.desc,) )
        else:
            self.result.append( "    -- FK reference from X%s_%s_%s -- %s" % (aSegment.message.name, aSegment.parent.name, aSegment.name, aSegment.desc,))
            raise StopDescent # already in a separate table definition
    def preLoop( self, aLoop, indent ):
        self.result.append( "    -- FK reference from X%s_%s -- %s" % (aLoop.message.name, aLoop.name, aLoop.desc,))
        raise StopDescent # already in a separate table definition

class SQLTableVisitor( StructureVisitor ):
    def __init__( self, *arg, **kw ):
        super( SQLTableVisitor, self ).__init__( *arg, **kw )
        self.result= []
    def preMessage( self, aMessage, indent ):
        """Table definitions of a Message."""
        self.result.append( "-- %s" % ( aMessage.desc,) )
        self.result.append( "CREATE TABLE X%s_%s(" % ( aMessage.message.name, aMessage.name ) )
        self.result.append( "    id INTEGER PRIMARY KEY AUTOINCREMENT," )
        for s in aMessage.structure:
            details= SQLDetailsVisitor( )
            s.visit( details )
            self.result.extend( details.result )
        self.result.append( "    xtra CHAR(8) -- placeholder")
        self.result.append( ") -- %s" % ( aMessage.name, ) )
        self.result.append( ";")
    def preLoop( self, aLoop, indent ):
        """Table definitions of a Loop."""
        self.result.append( "-- %s" % ( aLoop.desc,) )
        self.result.append( "CREATE TABLE X%s_%s(" % ( aLoop.message.name, aLoop.name ) )
        self.result.append( "    id INTEGER PRIMARY KEY AUTOINCREMENT," )
        self.result.append( "    loop INTEGER NOT NULL REFERENCES X%s_%s(id)," % ( aLoop.message.name, aLoop.parent.name) )
        for s in aLoop.structure:
            details= SQLDetailsVisitor( )
            s.visit( details )
            self.result.extend( details.result )
        self.result.append( "    xtra CHAR(8) -- placeholder")
        self.result.append( ") -- %s" % ( aLoop.name, ) )
        self.result.append( ";" )
    def preSegment( self, aSegment, indent ):
        if aSegment.repeat != "1":
            # Another table with an FK to this Segment
            name= "X%s_%s_%s%d" % ( aSegment.message.name, aSegment.parent.name, aSegment.name, aSegment.occurance )
            self.result.append( "-- %s" % ( aSegment.desc,) )
            self.result.append( "-- %r" % ( aSegment.props, ) )
            self.result.append( "CREATE TABLE %s(" % ( name, ) )
            self.result.append( "    id INTEGER PRIMARY KEY AUTOINCREMENT," )
            self.result.append( "    loop INTEGER NOT NULL REFERENCES X%s_%s(id)," % ( aSegment.message.name, aSegment.parent.name) )
            for s in aSegment.structure:
                details= SQLDetailsVisitor( )
                s.visit( details )
                self.result.extend( details.result )
            self.result.append( "    xtra CHAR(8) -- placeholder")
            self.result.append( ") -- %s" % ( aSegment.name, ) )
            self.result.append( ";" )
    def getSource( self ):
        return "\n".join( self.result )
